# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""Parse Pysa/Taint output for Zoncolan processing"""

import logging
import sys
from collections import defaultdict
from typing import (
    Any,
    Dict,
    IO,
    Iterable,
    List,
    NamedTuple,
    Optional,
    Set,
    Tuple,
    Union,
)

import ujson as json

from .. import errors
from ..analysis_output import AnalysisOutput, Metadata
from . import (
    ParseError,
    ParseConditionTuple,
    ParseFeature,
    ParseIssueConditionTuple,
    ParseIssueLeaf,
    ParseIssueTuple,
    ParsePosition,
    ParseTypeInterval,
    SourceLocation,
    flatten_features,
)
from .base_parser import (
    BaseParser,
    EntryPosition,
    ParseType,
    log_trace_keyerror_in_generator,
)

if sys.version_info >= (3, 8):
    from typing import TypedDict, Literal
else:
    from typing_extensions import TypedDict, Literal


log: logging.Logger = logging.getLogger("sapp")


class LeafWithPort(NamedTuple):
    name: Optional[str]
    kind: str
    port: Optional[str]


class LeafWithDistance(NamedTuple):
    name: Optional[str]
    kind: str
    distance: int


class LeafWithPortDistance(NamedTuple):
    name: Optional[str]
    kind: str
    port: Optional[str]
    distance: int

    def discard_port(self) -> LeafWithDistance:
        return LeafWithDistance(name=self.name, kind=self.kind, distance=self.distance)


class TraceFragment(TypedDict):
    callee: str
    port: str
    location: ParsePosition
    leaves: Iterable[LeafWithDistance]
    titos: Iterable[ParsePosition]
    features: Iterable[ParseFeature]
    type_interval: Optional[ParseTypeInterval]


class Parser(BaseParser):
    """The parser takes a json file as input, and provides a simplified output
    for the Processor.
    """

    _file_version: Optional[int] = None

    def parse(
        self, input: AnalysisOutput
    ) -> Iterable[Union[ParseConditionTuple, ParseIssueTuple]]:
        for handle in input.file_handles():
            for entry in self.parse_handle(handle):
                yield entry

    def parse_handle(
        self, handle: IO[str]
    ) -> Iterable[Union[ParseConditionTuple, ParseIssueTuple]]:
        for entry, _ in self._parse_entries(handle):
            yield from self._parse_by_type(entry)

    # Instead of returning the actual json from the AnalysisOutput, we return
    # location information so it can be retrieved later.
    def get_json_file_offsets(self, input: AnalysisOutput) -> Iterable[EntryPosition]:
        for handle in input.file_handles():
            for entry, position in self._parse_entries(handle):
                callable = self._get_callable(entry["data"].get("callable")).lstrip(
                    "\\"
                )
                yield EntryPosition(
                    callable=callable,
                    shard=position["shard"],
                    offset=position["offset"],
                )

    # Given a path and an offset, return the json in mostly-raw form.
    def get_json_from_file_offset(self, path: str, offset: int) -> Dict[str, Any]:
        with open(path) as fh:
            fh.seek(offset)
            return json.loads(fh.readline())

    def _parse_entries(
        self, handle: IO[str]
    ) -> Iterable[Tuple[Dict[str, Any], Dict[str, int]]]:
        """Parse analysis in jsonlines format:
        { "file_version": 2, "config": <json> }
        { <error1> }
        { <error2> }
        ...
        """
        file_version = self._parse_file_version(handle)
        if file_version < 2:
            raise ParseError(f"File version `{file_version}` is no longer supported.")
        if file_version > 3:
            raise ParseError(f"Unknown file version `{file_version}`.")
        self._file_version = file_version

        offset, line = handle.tell(), handle.readline()
        while line:
            entry = json.loads(line)
            if entry:
                position = {"shard": 0, "offset": offset}
                yield entry, position
            offset, line = handle.tell(), handle.readline()

    def _parse_file_version(self, handle: IO[str]) -> int:
        first_line = handle.readline().strip()
        try:
            json_first_line = json.loads(first_line)
            version = json_first_line["file_version"]
        except ValueError:
            raise ParseError("First line is not valid JSON.", received=first_line)
        except KeyError:
            raise ParseError(
                "First entry must have a `file_version` attribute.", received=first_line
            )

        return version

    def _parse_by_type(
        self, entry: Dict[str, Any]
    ) -> Iterable[Union[ParseConditionTuple, ParseIssueTuple]]:
        if entry["kind"] == "model":
            yield from self._parse_model(entry["data"])
        elif entry["kind"] == "issue":
            yield from self._parse_issue(entry["data"])
        else:
            raise ParseError("Unexpected kind in entry", received=entry["kind"])

    @staticmethod
    def _get_callable(callable: str) -> str:
        return callable

    @log_trace_keyerror_in_generator
    def _parse_model(self, json: Dict[str, Any]) -> Iterable[ParseConditionTuple]:
        callable = json["callable"]
        yield from self._parse_model_sources(callable, json.get("sources", []))
        yield from self._parse_model_sinks(callable, json.get("sinks", []))

    def _parse_model_sources(
        self, callable: str, source_traces: List[Dict[str, Any]]
    ) -> Iterable[ParseConditionTuple]:
        for source_trace in source_traces:
            port = source_trace["port"]
            for fragment in self._parse_trace_fragments(
                "source", source_trace["taint"]
            ):
                yield ParseConditionTuple(
                    type=ParseType.POSTCONDITION,
                    caller=callable,
                    callee=fragment["callee"],
                    callee_location=SourceLocation.from_typed_dict(
                        fragment["location"]
                    ),
                    filename=fragment["location"]["filename"],
                    titos=list(map(SourceLocation.from_typed_dict, fragment["titos"])),
                    leaves=[(leaf.kind, leaf.distance) for leaf in fragment["leaves"]],
                    caller_port=port,
                    callee_port=fragment["port"],
                    type_interval=None,
                    features=flatten_features(fragment["features"]),
                    annotations=[],
                )

    def _parse_model_sinks(
        self, callable: str, sink_traces: List[Dict[str, Any]]
    ) -> Iterable[ParseConditionTuple]:
        for sink_trace in sink_traces:
            port = sink_trace["port"]
            for fragment in self._parse_trace_fragments("sink", sink_trace["taint"]):
                yield ParseConditionTuple(
                    type=ParseType.PRECONDITION,
                    caller=callable,
                    callee=fragment["callee"],
                    callee_location=SourceLocation.from_typed_dict(
                        fragment["location"]
                    ),
                    filename=fragment["location"]["filename"],
                    titos=list(map(SourceLocation.from_typed_dict, fragment["titos"])),
                    leaves=[(leaf.kind, leaf.distance) for leaf in fragment["leaves"]],
                    caller_port=port,
                    callee_port=fragment["port"],
                    type_interval=None,
                    features=flatten_features(fragment["features"]),
                    annotations=[],
                )

    @log_trace_keyerror_in_generator
    def _parse_issue(self, json: Dict[str, Any]) -> Iterable[ParseIssueTuple]:
        (
            preconditions,
            final_sinks,
        ) = self._parse_issue_traces(json["traces"], "backward", "sink")
        (
            postconditions,
            initial_sources,
        ) = self._parse_issue_traces(json["traces"], "forward", "source")

        yield ParseIssueTuple(
            code=json["code"],
            line=json["line"],
            callable_line=json["callable_line"],
            start=json["start"],
            end=json["end"],
            callable=json["callable"],
            handle=self._generate_issue_master_handle(json),
            message=json["message"],
            filename=self._extract_filename(json["filename"]),
            preconditions=preconditions,
            final_sinks=final_sinks,
            postconditions=postconditions,
            initial_sources=initial_sources,
            fix_info=None,
            features=flatten_features(json["features"]),
        )

    def _generate_issue_master_handle(self, issue: Dict[str, Any]) -> str:
        line = issue["line"] - issue["callable_line"]
        return self.compute_master_handle(
            callable=issue["callable"],
            line=line,
            start=issue["start"],
            end=issue["end"],
            code=issue["code"],
        )

    def _extract_filename(self, complete_filename: str) -> str:
        repo_dirs = self.repo_dirs
        if not repo_dirs:
            return complete_filename
        if not complete_filename.startswith("/"):
            # already relative
            return complete_filename
        for repo_dir in repo_dirs:
            repo_dir = repo_dir.rstrip("/")
            if repo_dir != "" and complete_filename.startswith(repo_dir):
                return complete_filename[len(repo_dir) + 1 :]

        raise errors.AIException(
            "Expected filename ({}) to start with repo_dir ({}). "
            "Check the --repo-dir option.".format(complete_filename, repo_dirs)
        )

    def _parse_issue_traces(
        self,
        traces: List[Dict[str, Any]],
        name: Union[Literal["forward"], Literal["backward"]],
        leaf_port: Union[Literal["source"], Literal["sink"]],
    ) -> Tuple[List[ParseIssueConditionTuple], Set[ParseIssueLeaf]]:
        for trace in traces:
            if trace["name"] == name:
                return self._parse_issue_trace_fragments(leaf_port, trace["roots"])

        raise ParseError(f"Could not find `{name}` in trace.", received=traces)

    def _parse_issue_trace_fragments(
        self,
        leaf_port: Union[Literal["source"], Literal["sink"]],
        traces: List[Dict[str, Any]],
    ) -> Tuple[List[ParseIssueConditionTuple], Set[ParseIssueLeaf]]:
        fragments: List[ParseIssueConditionTuple] = []
        leaf_distances: Set[ParseIssueLeaf] = set()

        for trace in traces:
            for fragment in self._parse_trace_fragment(leaf_port, trace):
                leaves = fragment["leaves"]
                fragments.append(
                    ParseIssueConditionTuple(
                        callee=fragment["callee"],
                        port=fragment["port"],
                        location=SourceLocation.from_typed_dict(fragment["location"]),
                        leaves=[(leaf.kind, leaf.distance) for leaf in leaves],
                        titos=list(
                            map(SourceLocation.from_typed_dict, fragment["titos"])
                        ),
                        features=flatten_features(fragment["features"]),
                        type_interval=None,
                        annotations=[],
                    )
                )
                leaf_distances.update(
                    # pyre-fixme: Expected `str` but got `Optional[str]` for leaf.name
                    (leaf.name, leaf.kind, leaf.distance)
                    for leaf in leaves
                )

        return (fragments, leaf_distances)

    def _parse_trace_fragments(
        self,
        leaf_port: Union[Literal["source"], Literal["sink"]],
        traces: List[Dict[str, Any]],
    ) -> Iterable[TraceFragment]:
        for trace in traces:
            yield from self._parse_trace_fragment(leaf_port, trace)

    def _parse_trace_fragment(
        self,
        leaf_port: Union[Literal["source"], Literal["sink"]],
        trace: Dict[str, Any],
    ) -> Iterable[TraceFragment]:
        if self._file_version == 2:
            yield from self._parse_trace_fragment_v2(leaf_port, trace)
        elif self._file_version == 3:
            yield from self._parse_trace_fragment_v3(leaf_port, trace)
        else:
            raise ParseError("Unsupported file version", received=self._file_version)

    def _parse_trace_fragment_v2(
        self,
        leaf_port: Union[Literal["source"], Literal["sink"]],
        trace: Dict[str, Any],
    ) -> Iterable[TraceFragment]:
        # For now we don't have leaf distances.
        leaves = self._parse_leaves_v2(trace.get("leaves", []))
        if "root" in trace:
            leaf_name_and_port_to_leaves = defaultdict(list)
            for leaf in leaves:
                port = leaf.port or leaf_port
                callee_name = leaf.name or "leaf"
                leaf_name_and_port_to_leaves[(callee_name, port)].append(leaf)

            for ((callee_name, port), leaves) in leaf_name_and_port_to_leaves.items():
                fragment: TraceFragment = {
                    "callee": callee_name,
                    "port": port,
                    "location": self._adjust_location(trace["root"]),
                    "leaves": [
                        LeafWithDistance(name=leaf.name, kind=leaf.kind, distance=0)
                        for leaf in leaves
                    ],
                    "titos": list(map(self._adjust_location, trace.get("tito", []))),
                    "features": [],
                    "type_interval": None,
                }
                yield fragment
        elif "call" in trace:
            call = trace["call"]
            port = call["port"]
            resolves_to = call.get("resolves_to", [])
            length = call.get("length", 0)
            leaves = [
                LeafWithDistance(name=leaf.name, kind=leaf.kind, distance=length)
                for leaf in leaves
            ]

            for resolved in resolves_to:
                fragment: TraceFragment = {
                    "callee": resolved,
                    "port": port,
                    "location": self._adjust_location(call["position"]),
                    "leaves": leaves,
                    "titos": list(map(self._adjust_location, trace.get("tito", []))),
                    "features": [],
                    "type_interval": None,
                }
                yield fragment
        elif "decl" in trace:
            pass  # User-declared fragment.
        else:
            raise ParseError("Unexpected trace fragment.", received=trace)

    def _parse_trace_fragment_v3(
        self,
        leaf_port: Union[Literal["source"], Literal["sink"]],
        trace: Dict[str, Any],
    ) -> Iterable[TraceFragment]:
        tito_positions = list(map(self._adjust_location, trace.get("tito", [])))
        local_features = trace.get("local_features", [])

        if "root" in trace:
            location = self._adjust_location(trace["root"])

            # Turn leaves into direct callees and group by (callee, port)
            leaf_name_and_port_to_leaves: defaultdict[
                Tuple[str, str], List[LeafWithDistance]
            ] = defaultdict(list)
            for leaf in self._parse_leaves_v3(trace):
                callee_name = leaf.name or "leaf"
                callee_port = leaf.port or leaf_port
                leaf_name_and_port_to_leaves[(callee_name, callee_port)].append(
                    leaf.discard_port()
                )

            for (
                (callee_name, callee_port),
                leaves,
            ) in leaf_name_and_port_to_leaves.items():
                fragment: TraceFragment = {
                    "callee": callee_name,
                    "port": callee_port,
                    "location": location,
                    "leaves": leaves,
                    "titos": tito_positions,
                    "features": local_features,
                    "type_interval": None,
                }
                yield fragment
        elif "call" in trace:
            call = trace["call"]
            port = call["port"]
            location = self._adjust_location(call["position"])
            resolves_to = call.get("resolves_to", [])
            leaves: List[LeafWithDistance] = [
                leaf.discard_port() for leaf in self._parse_leaves_v3(trace)
            ]

            for resolved in resolves_to:
                fragment: TraceFragment = {
                    "callee": resolved,
                    "port": port,
                    "location": location,
                    "leaves": leaves,
                    "titos": tito_positions,
                    "features": local_features,
                    "type_interval": None,
                }
                yield fragment
        elif "decl" in trace:
            pass  # User-declared fragment.
        else:
            raise ParseError("Unexpected trace fragment.", received=trace)

    def _adjust_location(self, location: ParsePosition) -> ParsePosition:
        return {**location, "start": location["start"] + 1}  # pyre-ignore[7]

    def _parse_leaves_v2(self, leaves: List[Dict[str, Any]]) -> List[LeafWithPort]:
        return [
            LeafWithPort(
                name=leaf.get("name", None),
                kind=leaf["kind"],
                port=leaf.get("port", None),
            )
            for leaf in leaves
        ]

    def _parse_leaves_v3(self, trace: Dict[str, Any]) -> List[LeafWithPortDistance]:
        leaves: List[LeafWithPortDistance] = []
        for flow_details in trace.get("kinds", []):
            kind = flow_details["kind"]
            distance = flow_details.get("length", 0)
            for leaf in flow_details.get("leaves", [{}]):
                leaves.append(
                    LeafWithPortDistance(
                        name=leaf.get("name", None),
                        kind=kind,
                        port=leaf.get("port", None),
                        distance=distance,
                    )
                )
        return leaves

    @staticmethod
    def is_supported(metadata: Metadata) -> bool:
        return metadata.tool == "pysa"



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict


from typing import List, Optional, Tuple

from IPython.core.interactiveshell import InteractiveShell
from IPython.core.magic import Magics
from IPython.terminal.prompts import Prompts, Token
from prompt_toolkit import Application as CommandLineInterface
from pygments.token import _TokenType

from ..ui.interactive import Interactive


class CustomPrompt(Prompts, Magics):
    def in_prompt_tokens(
        self, cli: Optional[CommandLineInterface] = None
    ) -> List[Tuple[_TokenType, str]]:
        user_ns = self.shell.user_ns
        interactive: Interactive = user_ns[Interactive.SELF_SCOPE_KEY]
        tokens = []

        if interactive._current_run_id > 0:
            tokens.extend(
                [
                    (Token.Punctuation, "[ "),
                    (Token, f"run {interactive._current_run_id} "),
                ]
            )

        if interactive.current_issue_instance_id > 0:
            tokens.append((Token, f"> issue {interactive.current_issue_instance_id} "))
        elif interactive.current_frame_id > 0:
            tokens.append((Token, f"> frame {interactive.current_frame_id} "))

        current_callable = interactive.callable()
        if current_callable:
            tokens.append((Token.Name, f"> {current_callable} "))

        if len(tokens) > 0:
            tokens.append((Token.Punctuation, "]"))

        tokens.append((Token.Prompt, "\n>>> "))

        return tokens


def load_ipython_extension(ipython: InteractiveShell) -> None:
    # pyre-fixme[16]: `InteractiveShell` has no attribute `prompts`.
    ipython.prompts = CustomPrompt(ipython)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import logging
import os
from typing import Optional

import sqlalchemy

# pyre-fixme[21]: `flask` has no attribute `_app_ctx_stack`
from flask import Flask, send_from_directory, _app_ctx_stack
from flask.wrappers import Response
from flask_cors import CORS
from flask_graphql import GraphQLView
from pyre_extensions import none_throws
from sqlalchemy.orm import Session, scoped_session, sessionmaker

from .. import models
from ..db import DB
from .filters import ServeExportFilter
from .schema import schema

logging.basicConfig(
    format="%(asctime)s [%(levelname)s] %(message)s", level=logging.DEBUG
)
LOG: logging.Logger = logging.getLogger(__name__)


application = Flask(
    __name__, static_folder=os.path.join(os.path.dirname(__file__), "frontend", "build")
)

session: Optional[Session] = None


@application.teardown_request
def shutdown_session(exception: Optional[Exception] = None) -> None:
    if session is not None:
        # pyre-fixme[16]: `Session` has no attribute `remove`.
        session.remove()


@application.route("/", defaults={"path": ""})
@application.route("/<path:path>")
def serve(path: str) -> Response:
    LOG.info(f"Serving `{path}`...")
    static_folder = none_throws(application.static_folder)
    if path != "" and os.path.exists(static_folder + "/" + path):
        LOG.info("Found static resource.")
        return send_from_directory(static_folder, path)
    else:
        LOG.info("Resource not found. Falling back to `index.html`")
        return send_from_directory(static_folder, "index.html")


def start_server(
    database: DB,
    debug: bool,
    static_resources: Optional[str],
    source_directory: str,
    editor_schema: Optional[str],
) -> None:
    engine = sqlalchemy.create_engine(
        sqlalchemy.engine.url.URL("sqlite", database=database.dbname),
        echo=False,
        poolclass=None,
    )
    session = scoped_session(
        sessionmaker(bind=engine),
        # pyre-fixme[16]: `flask` has no attribute _app_ctx_stack
        scopefunc=_app_ctx_stack.__ident_func__,
    )
    # pyre-fixme[16]: `Type` has no attribute `query`.
    models.Base.query = session.query_property()
    # We have additional tables for the UI that need to be created.
    models.create(database)

    application.add_url_rule(
        "/graphql",
        view_func=GraphQLView.as_view(
            "graphql",
            schema=schema,
            graphiql=True,
            get_context=lambda: {
                "session": session,
                "source_directory": os.path.expanduser(source_directory),
                "editor_schema": editor_schema,
            },
        ),
    )
    application.add_url_rule(
        "/export_filter/<string:filter_name>",
        view_func=ServeExportFilter.as_view("filter_export_view", session=session),
    )
    if static_resources:
        application.static_folder = static_resources
    if debug:
        CORS(application, resources={r"/graphql": {"origins": "http://localhost:3000"}})
    application.run(debug=debug, host="localhost", port=5000)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import itertools


# pyre-fixme[3]: Return type must be annotated.
# pyre-fixme[2]: Parameter must be annotated.
# pyre-fixme[2]: Parameter must be annotated.
def split_every(n, iterable):
    """Yields batches of size 'n' from an iterable:

    list(split_every(2, range(10))) => [[0, 1], [2, 3], [4, 5], [6, 7], [8, 9]]
    """
    i = iter(iterable)
    piece = list(itertools.islice(i, n))
    while piece:
        yield piece
        piece = list(itertools.islice(i, n))



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import json
from dataclasses import dataclass, field
from json import JSONEncoder
from typing import Set, Dict, Union, List, Tuple

from sqlalchemy.orm import Session
from typing_extensions import TypeAlias

from .queries import get_warning_message_range
from .sarif_types import SARIFResult
from .ui.issues import IssueQueryResult


SARIFOutput: TypeAlias = Dict[
    str,
    Union[
        List[
            Dict[
                str,
                Union[
                    Dict[str, Dict[str, Union[List[Dict[str, str]], str]]],
                    List[SARIFResult],
                ],
            ]
        ],
        str,
    ],
]


@dataclass
class SARIF:
    version: str = "2.1.0"
    schema: str = "https://raw.githubusercontent.com/oasis-tcs/sarif-spec/master/Schemata/sarif-schema-2.1.0.json"  # noqa
    _tool_warning_code_ranges: Dict[str, Tuple[int, int]] = field(default_factory=dict)
    driver: Dict[str, Union[str, List[Dict[str, str]]]] = field(default_factory=dict)
    results: List[SARIFResult] = field(default_factory=list)

    def __init__(
        self, tool: str, session: Session, filtered_issues: Set[IssueQueryResult]
    ) -> None:
        self._tool_warning_code_ranges = {
            "mariana-trench": (4000, 5000),
            "pysa": (5000, 6000),
        }
        driver_json = {}
        if tool == "pysa":
            driver_json["name"] = "Pysa"
            driver_json["informationUri"] = "https://github.com/facebook/pyre-check/"

            tool_warning_messages = get_warning_message_range(
                session,
                self._tool_warning_code_ranges[tool][0],
                self._tool_warning_code_ranges[tool][1],
            )
            rules_json = []
            for rule in tool_warning_messages:
                rules_json.append({"id": str(rule.code), "name": rule.message})
            driver_json["rules"] = rules_json
        else:
            raise NotImplementedError

        self.driver = driver_json
        self.results = [issue.to_sarif() for issue in filtered_issues]

    def to_json(self, indent: int = 2) -> str:
        return json.dumps(self, cls=SARIFEncoder, indent=indent)


class SARIFEncoder(JSONEncoder):
    def default(self, o: SARIF) -> SARIFOutput:
        return {
            "version": o.version,
            "$schema": o.schema,
            "runs": [
                {
                    "tool": {"driver": o.driver},
                    "results": o.results,
                }
            ],
        }



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import List

import graphene
from sqlalchemy.orm import Session
from sqlalchemy.sql import func, distinct

from ..models import DBID, IssueInstance, Issue, IssueStatus, MetaRunToRunAssoc
from ..models import Run as RunColumn
from ..models import RunOrigin, RunStatus, TraceFrame


def latest(session: Session) -> DBID:
    return DBID(
        (
            session.query(func.max(RunColumn.id))
            .filter(RunColumn.status == RunStatus.FINISHED)
            .scalar()
        )
    )


class Run(graphene.ObjectType):
    run_id = graphene.ID()
    date = graphene.String()
    commit_hash = graphene.String()
    num_issues = graphene.Int()
    triaged_issues = graphene.Int()


def runs(session: Session) -> List[Run]:
    triaged_issues = (
        session.query(
            RunColumn.id.label("run_id"),
            func.count(distinct(IssueInstance.id)).label("count"),
        )
        .group_by(IssueInstance.run_id)
        .join(IssueInstance, IssueInstance.run_id == RunColumn.id)
        .join(Issue, Issue.id == IssueInstance.issue_id)
        .filter(Issue.status != IssueStatus.UNCATEGORIZED)
        .subquery()
    )
    return (
        session.query(
            RunColumn.id.label("run_id"),
            RunColumn.date,
            RunColumn.commit_hash,
            func.count(distinct(IssueInstance.id)).label("num_issues"),
            triaged_issues.c.count.label("triaged_issues"),
        )
        .group_by(RunColumn)
        .join(IssueInstance, IssueInstance.run_id == RunColumn.id, isouter=True)
        .join(triaged_issues, triaged_issues.c.run_id == RunColumn.id, isouter=True)
        .filter(RunColumn.status == "finished")
        .order_by(RunColumn.id.desc())
        .all()
    )


class EmptyDeletionError(Exception):
    pass


def delete_run(session: Session, id: str) -> None:
    deleted_run_rows = session.query(RunColumn).filter(RunColumn.id == id).delete()
    if deleted_run_rows == 0:
        raise EmptyDeletionError(f'No run with `id` "{id}" exists.')
    session.query(IssueInstance).filter(IssueInstance.run_id == id).delete()
    session.query(TraceFrame).filter(TraceFrame.run_id == id).delete()
    session.query(RunOrigin).filter(RunOrigin.run_id == id).delete()
    session.query(MetaRunToRunAssoc).filter(MetaRunToRunAssoc.run_id == id).delete()
    session.commit()



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""Abstract Parser for Zoncolan like output"""

import logging
import pprint
from collections import defaultdict
from typing import (
    Any,
    Dict,
    Iterable,
    List,
    NamedTuple,
    Set,
    TextIO,
    Tuple,
    Union,
    cast,
)

import xxhash

from ..analysis_output import AnalysisOutput, Metadata
from . import (
    DictEntries,
    DictKey,
    Optional,
    ParseCondition,
    ParseConditionTuple,
    ParseIssue,
    ParseIssueTuple,
    ParseType,
    PipelineStep,
    Summary,
)

# if these imports have the same name we get a linter error
try:
    import ujson as json
except ImportError:
    import json  # noqa


log: logging.Logger = logging.getLogger("sapp")


# The callable's json output can be found at the given sharded file and offset.
# Used for debugging.
class EntryPosition(NamedTuple):
    callable: str
    shard: int
    offset: int


# pyre-ignore[2]
# pyre-ignore[3]
def log_trace_keyerror(func):
    # pyre-ignore[2]
    # pyre-ignore[3]
    # pyre-ignore[53]
    def wrapper(self, json, *args):
        try:
            return func(self, json, *args)
        except KeyError:
            # The most common problem with parsing json is not finding
            # a field you expect, so we'll catch those and log them, but move
            # on.
            log.exception(
                "Unable to parse trace for the following:\n%s", pprint.pformat(json)
            )
            return ([], {})

    return wrapper


# pyre-ignore[2]
# pyre-ignore[3]
def log_trace_keyerror_in_generator(func):
    # pyre-ignore[2]
    # pyre-ignore[3]
    # pyre-ignore[53]
    def wrapper(self, json, *args):
        try:
            yield from func(self, json, *args)
        except KeyError:
            # The most common problem with parsing json is not finding
            # a field you expect, so we'll catch those and log them, but move
            # on.
            log.exception(
                "Unable to parse trace for the following:\n%s", pprint.pformat(json)
            )
            return
            yield

    return wrapper


class BaseParser(PipelineStep[AnalysisOutput, DictEntries]):
    """The parser takes a json file as input, and provides a simplified output
    for the Processor.
    """

    def __init__(self, repo_dirs: Optional[List[str]] = None) -> None:
        """
        repo_dirs: Possible absolute paths analyzed during the run. This is used to relativize
        paths in the input. These paths are NOT guaranteed to exist on the current machine disk!
        """
        self.repo_dirs: List[str] = repo_dirs or []

    def initialize(self, metadata: Optional[Metadata]) -> None:
        return

    # @abstractmethod
    def parse(
        self, input: AnalysisOutput
    ) -> Iterable[
        Union[ParseCondition, ParseIssue, ParseConditionTuple, ParseIssueTuple]
    ]:
        """Must return objects with a 'type': ParseType field."""
        raise NotImplementedError("Abstract method called!")

    # @abstractmethod
    def parse_handle(
        self, handle: TextIO
    ) -> Iterable[
        Union[ParseCondition, ParseIssue, ParseConditionTuple, ParseIssueTuple]
    ]:
        """Must return objects with a 'type': ParseType field."""
        raise NotImplementedError("Abstract method called!")

    def _analysis_output_to_parsed_tuples(
        self, input: AnalysisOutput
    ) -> Iterable[
        Tuple[ParseType, DictKey, Union[ParseConditionTuple, ParseIssueTuple]]
    ]:
        entries = self.parse(input)

        for e in entries:
            if isinstance(e, ParseConditionTuple):
                typ = e.type
                key = (e.caller, e.caller_port)
            elif isinstance(e, ParseIssueTuple):
                typ = ParseType.ISSUE
                key = e.handle
            else:
                # legacy raw dicts
                typ = e["type"]
                if typ == ParseType.ISSUE:
                    e = ParseIssueTuple.from_typed_dict(cast(ParseIssue, e))
                    key = e.handle
                elif typ == ParseType.PRECONDITION or typ == ParseType.POSTCONDITION:
                    e = ParseConditionTuple.from_typed_dict(cast(ParseCondition, e))
                    key = (e.caller, e.caller_port)
                else:
                    raise Exception("Unknown ParseType")
            yield typ, key, e

    def analysis_output_to_dict_entries(
        self,
        inputfile: AnalysisOutput,
        previous_issue_handles: Optional[AnalysisOutput],
        linemapfile: Optional[str],
    ) -> DictEntries:
        """Here we take input generators and return a dict with issues,
        preconditions, and postconditions separated. If there is only a single
        generator file, it's simple. If we also pass in a generator from a
        previous inputfile then there are a couple extra steps:

        1. If an issue was seen in the previous inputfile then we won't return
        it, because it's not new.
        2. In addition, we take an optional linemap file that maps for each
        filename, each new file line position to a list of old file line
        position. This is used to adjust handles to we can recognize when issues
        moved.
        """

        issues: List[ParseIssueTuple] = []
        previous_handles: Set[str] = set()
        conditions: Dict[ParseType, Dict[DictKey, List[ParseConditionTuple]]] = {
            ParseType.PRECONDITION: defaultdict(list),
            ParseType.POSTCONDITION: defaultdict(list),
        }

        # If we have a mapfile, create the map.
        if linemapfile:
            log.info("Parsing linemap file")
            with open(linemapfile, "r") as f:
                linemap = json.load(f)
        else:
            linemap = None

        # Save entry info from the parent analysis, if there is one.
        if previous_issue_handles:
            log.info("Parsing previous issue handles")
            for f in previous_issue_handles.file_handles():
                handles = f.read().splitlines()
                previous_handles = set(filter(lambda h: not h.startswith("#"), handles))

        log.info("Parsing analysis output...")
        for typ, key, e in self._analysis_output_to_parsed_tuples(inputfile):
            if typ == ParseType.ISSUE:
                e = cast(ParseIssueTuple, e)
                # We are only interested in issues that weren't in the previous
                # analysis.
                if not self._is_existing_issue(linemap, previous_handles, e, key):
                    issues.append(e.interned())
            elif typ == ParseType.PRECONDITION or typ == ParseType.POSTCONDITION:
                e = cast(ParseConditionTuple, e)
                conditions[typ][key].append(e.interned())

        return {
            "issues": issues,
            "preconditions": conditions[ParseType.PRECONDITION],
            "postconditions": conditions[ParseType.POSTCONDITION],
        }

    def _is_existing_issue(
        self,
        linemap: Dict[str, Any],
        old_handles: Set[str],
        new_issue: ParseIssueTuple,
        new_handle: DictKey,
    ) -> bool:
        if new_handle in old_handles:
            return True
        if not linemap:
            return False
        filename = new_issue.filename
        old_map = linemap.get(filename, {})
        old_lines = old_map.get(str(new_issue.line), [])
        # Once this works, we should remove the "relative" line from the handle
        # and use the absolute one to avoid having to map both the start of the
        # method and the line in the method.

        # Consider all possible old lines
        for old_line in old_lines:
            old_handle = BaseParser.compute_diff_handle(
                filename, old_line, new_issue.code
            )
            if old_handle in old_handles:
                return True
        return False

    def run(
        self, input: AnalysisOutput, summary: Summary
    ) -> Tuple[DictEntries, Summary]:
        return (
            self.analysis_output_to_dict_entries(
                input,
                summary.get("previous_issue_handles"),
                summary.get("old_linemap_file"),
            ),
            summary,
        )

    @staticmethod
    def compute_master_handle(
        callable: str, line: int, start: int, end: int, code: int
    ) -> str:
        key = "{callable}:{line}|{start}|{end}:{code}".format(
            callable=callable, line=line, start=start, end=end, code=code
        )
        return BaseParser.compute_handle_from_key(key)

    @staticmethod
    def compute_diff_handle(filename: str, old_line: int, code: int) -> str:
        """Uses the absolute line and ignores the callable/character offsets.
        Used only in determining whether new issues are old issues.
        """
        key = "{filename}:{old_line}:{code}".format(
            filename=filename, old_line=old_line, code=code
        )
        return BaseParser.compute_handle_from_key(key)

    @staticmethod
    def compute_handle_from_key(key: str) -> str:
        hash_gen = xxhash.xxh64()
        hash_gen.update(key.encode())
        hash_ = hash_gen.hexdigest()
        return key[: 255 - len(hash_) - 1] + ":" + hash_

    @staticmethod
    def is_supported(metadata: Metadata) -> bool:
        raise NotImplementedError("Subclasses should implement this!")

    # Instead of returning the actual json from the AnalysisOutput, we return
    # location information so it can be retrieved later.
    def get_json_file_offsets(self, input: AnalysisOutput) -> Iterable[EntryPosition]:
        raise NotImplementedError("get_json_file_offset not implemented")

    # Given a path and an offset, return the json in mostly-raw form.
    def get_json_from_file_offset(
        self, path: str, offset: int
    ) -> Optional[Dict[str, Any]]:
        raise NotImplementedError("get_json_from_file_offset not implemented")



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Set, Tuple

from . import DictEntries, ParseIssueTuple, PipelineStep, Summary


class WarningCodeFilter(PipelineStep[DictEntries, DictEntries]):
    def __init__(self, codes_to_keep: Set[int]) -> None:
        self.codes_to_keep: Set[int] = codes_to_keep

    def _should_skip_issue(self, issue: ParseIssueTuple) -> bool:
        return issue.code not in self.codes_to_keep

    def run(self, input: DictEntries, summary: Summary) -> Tuple[DictEntries, Summary]:
        filtered_issues = []
        for issue in input["issues"]:
            if self._should_skip_issue(issue):
                continue
            filtered_issues.append(issue)

        input["issues"] = filtered_issues

        return input, summary



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict
from typing import List, Optional

from sqlalchemy.orm import Session, Query
from sqlalchemy.sql import func

from .decorators import UserError
from .models import (
    DBID,
    IssueInstance,
    IssueInstanceSharedTextAssoc,
    Run,
    RunStatus,
    WarningMessage,
    SharedText,
    SharedTextKind,
)


def get_warning_message_range(
    session: Session, startingCode: int, endingCode: int
) -> List[WarningMessage]:
    return (
        session.query(WarningMessage)
        .filter(WarningMessage.code >= startingCode)
        .filter(WarningMessage.code < endingCode)
        .all()
    )


def get_warning_message(
    session: Session,
    code: int,
) -> Optional[WarningMessage]:
    return (
        session.query(WarningMessage).filter(WarningMessage.code == code).one_or_none()
    )


def latest_run_id(
    session: Session,
) -> DBID:
    return (
        session.query(func.max(Run.id))
        .filter(Run.status == RunStatus.FINISHED)
        .scalar()
    )


def _leaf_detail_kind(kind: str) -> SharedTextKind:
    """
    Convert a raw kind (source|sink|feature) into a SharedTextKind,
    converting source and sink into detail kinds.
    """
    text_kind = SharedTextKind.from_string(kind)
    if text_kind is None:
        raise UserError(f"Invalid kind {kind}")
    if text_kind == SharedTextKind.source:
        text_kind = SharedTextKind.source_detail
    elif text_kind == SharedTextKind.sink:
        text_kind = SharedTextKind.sink_detail
    return text_kind


def leaves(
    session: Session,
    kind: str,
    run_id: DBID,
) -> Query:
    text_kind = _leaf_detail_kind(kind)
    return (
        session.query(IssueInstanceSharedTextAssoc.shared_text_id, SharedText.contents)
        .join(
            SharedText,
            IssueInstanceSharedTextAssoc.shared_text_id == SharedText.id,
        )
        .join(
            IssueInstance,
            IssueInstanceSharedTextAssoc.issue_instance_id == IssueInstance.id,
        )
        .filter(IssueInstance.run_id == run_id)
        .filter(SharedText.kind == text_kind)
    )



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-unsafe

from __future__ import annotations

import enum
import logging
from datetime import datetime
from decimal import Decimal
from itertools import islice
from typing import Any, Dict, List, NamedTuple, Optional, Set, Type

from graphene_sqlalchemy.converter import (
    convert_column_to_int_or_id,
    convert_column_to_string,
    convert_sqlalchemy_type,
)
from sqlalchemy import (
    Boolean,
    Column,
    DateTime,
    Enum,
    Float,
    Index,
    Integer,
    String,
    func,
    types,
)
from sqlalchemy.dialects.mysql import BIGINT
from sqlalchemy.exc import NoSuchTableError
from sqlalchemy.ext.associationproxy import association_proxy
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import Session, relationship

from .db import DB
from .db_support import (
    DBID,
    BIGDBIDType,
    DBIDType,
    MutableRecordMixin,
    PrepareMixin,
    PrimaryKeyBase,
    PrimaryKeyGeneratorBase,
    RecordMixin,
)
from .decorators import classproperty
from .pipeline import SourceLocation

log: logging.Logger = logging.getLogger("sapp")


Base = declarative_base()
INNODB_MAX_INDEX_LENGTH = 767
HANDLE_LENGTH = 255
MESSAGE_LENGTH = 4096
SHARED_TEXT_LENGTH = 4096

"""Models used to represent DB entries

An Issue is a particular problem found. It can exist across multiple commits.  A
Run is a single run of Zoncolan over a specific commit. It may find new Issues,
or existing Issues.  Each run is tied to Issues through IssueInstances.
IssueInstances have per run information, like source location, while Issues have
attributes like the status of an issue.
"""


class LeafMapping(NamedTuple):
    caller_leaf: int
    callee_leaf: int
    transform: int


class SourceLocationType(types.TypeDecorator):
    """Defines a new type of SQLAlchemy to store source locations.

    In python land we use SourceLocation, but when stored in the databae we just
    split the fields with |
    """

    impl = types.String
    cache_ok = False

    def __init__(self) -> None:
        super(SourceLocationType, self).__init__(length=255)

    def process_bind_param(self, value, dialect):
        """
        SQLAlchemy uses this to convert a SourceLocation object into a string.
        """
        if value is None:
            return None
        return SourceLocation.to_string(value)

    def process_result_value(self, value, dialect) -> Optional[SourceLocation]:
        """
        SQLAlchemy uses this to convert a string into a SourceLocation object.
        We separate the fields by a |
        """
        if value is None:
            return None

        p = value.split("|")

        if len(p) == 0:
            return None
        return SourceLocation.of(*map(int, p))


class SourceLocationsType(types.TypeDecorator):
    """Defines a type to store multiple source locations in a single string"""

    impl = types.String
    cache_ok = False

    def __init__(self) -> None:
        super(SourceLocationsType, self).__init__(length=4096)

    def process_bind_param(self, value, dialect) -> Optional[str]:
        if value is None:
            return None
        return ",".join([SourceLocation.to_string(location) for location in value])

    def process_result_value(self, value, dialect):
        if value is None or value == "":
            return []
        assert isinstance(value, str), "Invalid SourceLocationsType %s" % str(value)
        locations = value.split(",")
        return [SourceLocation.from_string(location) for location in locations]


# See Issue.merge for information about replace_assocs


class IssueDBID(DBID):
    __slots__ = ["replace_assocs"]

    def __init__(self, id=None) -> None:
        super().__init__(id)
        self.replace_assocs = False


class IssueDBIDType(DBIDType):
    def process_result_value(self, value, dialect) -> IssueDBID:
        return IssueDBID(value)


class IssueBIGDBIDType(BIGDBIDType):
    def process_result_value(self, value, dialect) -> IssueDBID:
        return IssueDBID(value)


class IssueInstanceTraceFrameAssoc(Base, PrepareMixin, RecordMixin):

    __tablename__ = "issue_instance_trace_frame_assoc"

    issue_instance_id = Column(
        "issue_instance_id", BIGDBIDType, primary_key=True, nullable=False
    )

    trace_frame_id = Column(
        "trace_frame_id", BIGDBIDType, primary_key=True, nullable=False, index=True
    )

    issue_instance = relationship(
        "IssueInstance",
        primaryjoin=(
            "IssueInstanceTraceFrameAssoc.issue_instance_id == "
            "foreign(IssueInstance.id)"
        ),
        uselist=False,
    )

    trace_frame = relationship(
        "TraceFrame",
        primaryjoin=(
            "IssueInstanceTraceFrameAssoc.trace_frame_id == foreign(TraceFrame.id)"
        ),
        uselist=False,
    )

    @classmethod
    def merge(cls, session, items):
        return cls._merge_assocs(
            session, items, cls.issue_instance_id, cls.trace_frame_id
        )


class SharedTextKind(enum.Enum):
    # Do NOT reorder the enums. Depending on the type of database, existing
    # DBs may have these enums represented internally as ints based on the
    # order shown here, and changing it here messes up existing data. This
    # also means that new enums should be added AT THE END of the list.
    feature = enum.auto()
    message = enum.auto()
    source = enum.auto()
    sink = enum.auto()
    callable = enum.auto()
    filename = enum.auto()
    source_detail = enum.auto()
    sink_detail = enum.auto()

    @classproperty
    def FEATURE(cls):  # noqa
        return cls.feature

    @classproperty
    def MESSAGE(cls):  # noqa
        return cls.message

    @classproperty
    def SOURCE(cls):  # noqa
        return cls.source

    @classproperty
    def SINK(cls):  # noqa
        return cls.sink

    @classproperty
    def CALLABLE(cls):  # noqa
        return cls.callable

    @classproperty
    def FILENAME(cls):  # noqa
        return cls.filename

    @classproperty
    def SOURCE_DETAIL(cls):  # noqa
        return cls.source_detail

    @classproperty
    def SINK_DETAIL(cls):  # noqa
        return cls.sink_detail

    @classmethod
    def from_string(cls, string: str) -> Optional[SharedTextKind]:
        return cls.__members__.get(string)


class SharedText(Base, PrepareMixin, RecordMixin):
    """Any string-ish type that can be shared as a property of some other
    object. (e.g. features, sources, sinks). The table name 'messages' is due
    to legacy reasons."""

    __tablename__ = "messages"

    __table_args__ = (Index("ix_messages_handle", "contents", "kind"),)

    # pyre-fixme[8]: Attribute has type `DBID`; used as `Column[typing.Any]`.
    id: DBID = Column(BIGDBIDType, primary_key=True)

    # pyre-fixme[8]: Attribute has type `str`; used as `Column[str]`.
    contents: str = Column(
        String(length=SHARED_TEXT_LENGTH), nullable=False, index=True
    )

    # pyre-fixme[8]: Attribute has type `SharedTextKind`; used as `Column[str]`.
    kind: SharedTextKind = Column(
        Enum(SharedTextKind), server_default="feature", nullable=False, index=True
    )

    issue_instances = association_proxy("shared_text_issue_instance", "issue_instance")

    shared_text_issue_instance = relationship(
        "IssueInstanceSharedTextAssoc",
        primaryjoin=(
            "SharedText.id == foreign(IssueInstanceSharedTextAssoc.shared_text_id)"
        ),
    )

    trace_frames = association_proxy("shared_text_trace_frame", "trace_frames")

    shared_text_trace_frame = relationship(
        "TraceFrameLeafAssoc",
        primaryjoin=("SharedText.id == foreign(TraceFrameLeafAssoc.leaf_id)"),
    )

    @classmethod
    def merge(cls, session, items):
        return cls._merge_by_keys(
            session,
            items,
            lambda item: "%s:%s" % (item.contents, item.kind),
            cls.contents,
            cls.kind,
        )


class IssueInstanceSharedTextAssoc(Base, PrepareMixin, RecordMixin):
    """Assoc table between issue instances and its properties that are
    representable by a string. The DB table name and column names are due to
    legacy reasons and warrant some explanation:
    - 'Features' used to be the only shared text of the assoc, now, the assoc
      also accounts for 'Sources' and 'Sinks' and possibly more.
    - 'messages' table used to be only for 'messages', now, it contains
      features, sources and sinks and possibly more.
    - It is expensive to rename the DB tables, so renaming only happened in
      the model. This is why it looks like we have 3 different terms for the
      same thing: 'messages', 'shared_text', 'features'.

    When in doubt, trust the property and method names used in the model and
    refer to the relationship joins for how objects relate to each other.
    """

    __tablename__ = "issue_instance_feature_assoc"

    issue_instance_id = Column(
        "issue_instance_id", BIGDBIDType, primary_key=True, nullable=False
    )

    shared_text_id = Column("feature_id", BIGDBIDType, primary_key=True, nullable=False)

    issue_instance = relationship(
        "IssueInstance",
        primaryjoin=(
            "IssueInstanceSharedTextAssoc.issue_instance_id =="
            "foreign(IssueInstance.id)"
        ),
        uselist=False,
    )

    shared_text = relationship(
        "SharedText",
        primaryjoin=(
            "IssueInstanceSharedTextAssoc.shared_text_id == foreign(SharedText.id)"
        ),
        uselist=False,
    )

    @classmethod
    def merge(cls, session, items):
        return cls._merge_assocs(
            session, items, cls.issue_instance_id, cls.shared_text_id
        )


class TraceKind(enum.Enum):
    # Do NOT reorder the enums. Depending on the type of database, existing
    # DBs may have these enums represented internally as ints based on the
    # order shown here, and changing it here messes up existing data. This
    # also means that new enums should be added AT THE END of the list.
    precondition = enum.auto()
    postcondition = enum.auto()

    @classproperty
    def PRECONDITION(cls):  # noqa
        return cls.precondition

    @classproperty
    def POSTCONDITION(cls):  # noqa
        return cls.postcondition

    @classmethod
    def create_from_string(cls, value: str) -> TraceKind:
        if value == "precondition":
            return cls.precondition
        if value == "postcondition":
            return cls.postcondition
        raise ValueError(f"`{value}` is not a valid `TraceKind`")


class IssueInstance(Base, PrepareMixin, MutableRecordMixin):
    """A particularly instance of an issue found in a run"""

    __tablename__ = "issue_instances"

    # pyre-fixme[8]: Attribute has type `DBID`; used as `Column[typing.Any]`.
    id: DBID = Column(BIGDBIDType, primary_key=True)

    location = Column(
        SourceLocationType,
        nullable=False,
        doc="Location (possibly a range) of the issue",
    )

    filename_id = Column(BIGDBIDType, nullable=False, server_default="0", default=0)

    filename = relationship(
        "SharedText",
        primaryjoin="foreign(SharedText.id) == IssueInstance.filename_id",
        uselist=False,
    )

    callable_id = Column(BIGDBIDType, nullable=False, server_default="0", default=0)

    callable = relationship(
        "SharedText",
        primaryjoin="foreign(SharedText.id) == IssueInstance.callable_id",
        uselist=False,
    )

    is_new_issue: Column[Optional[bool]] = Column(
        Boolean,
        index=True,
        default=False,
        doc="True if the issue did not exist before this instance",
    )

    run_id = Column(BIGDBIDType, nullable=False, index=True)

    issue_id = Column(BIGDBIDType, nullable=False, index=True)

    issue = relationship(
        "Issue",
        primaryjoin="foreign(Issue.id) == IssueInstance.issue_id",
        uselist=False,
    )

    fix_info_id = Column(BIGDBIDType, nullable=True)

    fix_info = relationship(
        "IssueInstanceFixInfo",
        primaryjoin=("foreign(IssueInstanceFixInfo.id) == IssueInstance.fix_info_id"),
        uselist=False,
    )

    message_id = Column(BIGDBIDType, nullable=True)

    message = relationship(
        "SharedText",
        primaryjoin="foreign(SharedText.id) == IssueInstance.message_id",
        uselist=False,
    )

    trace_frames = association_proxy("issue_instance_trace_frame", "trace_frame")

    issue_instance_trace_frame = relationship(
        "IssueInstanceTraceFrameAssoc",
        primaryjoin=(
            "IssueInstance.id == "
            "foreign(IssueInstanceTraceFrameAssoc.issue_instance_id)"
        ),
    )

    shared_texts = association_proxy("issue_instance_shared_text", "shared_text")

    issue_instance_shared_text = relationship(
        "IssueInstanceSharedTextAssoc",
        primaryjoin=(
            "IssueInstance.id == "
            "foreign(IssueInstanceSharedTextAssoc.issue_instance_id)"
        ),
    )

    min_trace_length_to_sources: Column[Optional[int]] = Column(
        Integer, nullable=True, doc="The minimum trace length to sources"
    )

    min_trace_length_to_sinks: Column[Optional[int]] = Column(
        Integer, nullable=True, doc="The minimum trace length to sinks"
    )

    rank: Column[Optional[int]] = Column(
        Integer,
        server_default="0",
        doc="The higher the rank, the higher the priority for this issue",
    )

    callable_count: Column[Optional[int]] = Column(
        Integer,
        server_default="0",
        doc="Number of issues in this callable for this run",
    )

    min_trace_length_to_entrypoints: Column[Optional[int]] = Column(
        Integer, nullable=True, doc="The minimum trace length to entrypoints"
    )

    def get_shared_texts_by_kind(self, kind: SharedTextKind) -> List[SharedText]:
        return [text for text in self.shared_texts if text.kind == kind]

    def get_trace_frames_by_kind(self, kind: TraceKind):
        return [frame for frame in self.trace_frames if frame.kind == kind]

    @classmethod
    def merge(cls, session, items):
        for i in items:
            # If the issue is new, then the instance has to be new. But note
            # that we still may need RunDiffer, because issues that disappeared
            # for a while and then came back are also marked new.
            i.is_new_issue = i.issue_id.is_new
            yield i


class IssueStatus(enum.Enum):
    """Issues are born uncategorized. Humans can
    set it to FALSE_POSITIVE or VALID_BUG upon review."""

    # Do NOT reorder the enums. Depending on the type of database, existing
    # DBs may have these enums represented internally as ints based on the
    # order shown here, and changing it here messes up existing data. This
    # also means that new enums should be added AT THE END of the list.
    """An issue that hasn't been marked as a bug or FP"""
    uncategorized = enum.auto()

    """Not a security bug, but a bad practice. Still needs fixing."""
    bad_practice = enum.auto()

    """False positive from analysis"""
    false_positive = enum.auto()

    """Reviewed and seen to be a valid bug that needs fixing"""
    valid_bug = enum.auto()

    """I don't care about this particular issue,
    but still want to see issues of this kind."""
    do_not_care = enum.auto()

    @classproperty
    def UNCATEGORIZED(cls):  # noqa
        return cls.uncategorized

    @classproperty
    def BAD_PRACTICE(cls):  # noqa
        return cls.bad_practice

    @classproperty
    def FALSE_POSITIVE(cls):  # noqa
        return cls.false_positive

    @classproperty
    def VALID_BUG(cls):  # noqa
        return cls.valid_bug

    @classproperty
    def DO_NOT_CARE(cls):  # noqa
        return cls.do_not_care


class Issue(Base, PrepareMixin, MutableRecordMixin):
    """An issue coming from the static analysis.

    An issue can persist across multiple runs, even if it moves around in the
    code.
    """

    __tablename__ = "issues"

    # pyre-fixme[8]: Attribute has type `IssueDBID`; used as `Column[typing.Any]`.
    id: IssueDBID = Column(IssueBIGDBIDType, primary_key=True, nullable=False)

    handle: Column[str] = Column(
        String(length=HANDLE_LENGTH),
        nullable=False,
        unique=True,
        doc="This handle should uniquely identify an issue across runs on "
        + "different code revisions",
    )

    code: Column[int] = Column(
        Integer, doc="Code identifiying the issue type", nullable=False, index=True
    )

    instances = relationship(
        "IssueInstance", primaryjoin="Issue.id == foreign(IssueInstance.issue_id)"
    )

    first_seen: Column[datetime] = Column(
        DateTime,
        doc="time of the first run that found this issue",
        nullable=False,
        index=True,
    )

    status: Column[str] = Column(
        Enum(IssueStatus),
        doc="Shows the issue status from the latest run",
        server_default="uncategorized",
        nullable=False,
        index=True,
    )

    task_number: Column[Optional[int]] = Column(
        Integer, doc="Task number (not fbid) that is tracking this issue"
    )

    triage_history_fbid: Column[Optional[int]] = Column(
        BIGINT(unsigned=True),
        nullable=True,
        doc="FBID for EntZoncolanIssueTriageHistory",
    )

    feedback_fbid: Column[Optional[int]] = Column(
        BIGINT(unsigned=True), nullable=True, doc="FBID for EntZoncolanFeedback"
    )

    @classmethod
    def _take(cls, n, iterable):
        "Return first n items of the iterable as a list"
        return list(islice(iterable, n))

    @classmethod
    def merge(cls, session, issues):
        return cls._merge_by_key(session, issues, cls.handle)


class RunStatus(enum.Enum):
    # Do NOT reorder the enums. Depending on the type of database, existing
    # DBs may have these enums represented internally as ints based on the
    # order shown here, and changing it here messes up existing data. This
    # also means that new enums should be added AT THE END of the list.
    finished = enum.auto()
    incomplete = enum.auto()
    skipped = enum.auto()
    failed = enum.auto()

    @classproperty
    def FINISHED(cls):  # noqa
        return cls.finished

    @classproperty
    def INCOMPLETE(cls):  # noqa
        return cls.incomplete

    @classproperty
    def SKIPPED(cls):  # noqa
        return cls.skipped

    @classproperty
    def FAILED(cls):  # noqa
        return cls.failed


class PurgeStatus(enum.Enum):
    "Purge status of a run"

    # Do NOT reorder the enums. Depending on the type of database, existing
    # DBs may have these enums represented internally as ints based on the
    # order shown here, and changing it here messes up existing data. This
    # also means that new enums should be added AT THE END of the list.

    # Run has not been touched by purging automation
    unpurged = enum.auto()
    # Issue instances associated with an untriaged issue have been deleted
    # Trace frames marked as UNREACHABLE have been deleted
    purged = enum.auto()
    # Trace frames not reachable by an issue instance whose issue is triaged have been
    # marked UNREACHABLE.
    ready_to_purge = enum.auto()

    @classproperty
    def UNPURGED(cls) -> str:  # noqa
        # pyre-ignore[7]: Coerce to string for SQLAlchemy
        return cls.unpurged

    @classproperty
    def READY_TO_PURGE(cls) -> str:  # noqa
        # pyre-ignore[7]: Coerce to string for SQLAlchemy
        return cls.ready_to_purge

    @classproperty
    def PURGED(cls) -> str:  # noqa
        # pyre-ignore[7]: Coerce to string for SQLAlchemy
        return cls.purged


class FrameReachability(enum.Enum):
    "Internal reachability status of a trace frame"

    # Do NOT reorder the enums. Depending on the type of database, existing
    # DBs may have these enums represented internally as ints based on the
    # order shown here, and changing it here messes up existing data. This
    # also means that new enums should be added AT THE END of the list.
    unreachable = enum.auto()
    reachable = enum.auto()

    @classproperty
    def UNREACHABLE(cls) -> str:  # noqa
        # pyre-ignore[7]: Coerce to string for SQLAlchemy
        return cls.unreachable

    @classproperty
    def REACHABLE(cls) -> str:  # noqa
        # pyre-ignore[7]: Coerce to string for SQLAlchemy
        return cls.reachable


CURRENT_DB_VERSION = 1


class Run(Base):
    """A particular run of the static analyzer.

    Each time output is parsed from the static analyzer we generate a new run. A
    run has multiple IssueInstances."""

    __tablename__ = "runs"

    id = Column(BIGDBIDType, primary_key=True)

    job_id: Column[Optional[str]] = Column(String(length=255), index=True)

    date: Column[datetime] = Column(
        DateTime, doc="The date/time the analysis was run", nullable=False
    )

    commit_hash: Column[Optional[str]] = Column(
        String(length=255),
        doc="The commit hash of the codebase",
        nullable=True,
        index=True,
    )

    revision_id: Column[Optional[int]] = Column(
        Integer, doc="Phabricator Diff number (DXXXXXX)", nullable=True, index=True
    )

    differential_id: Column[Optional[int]] = Column(
        Integer,
        doc="Phabricator Version number",
        nullable=True,
        index=True,
    )

    hh_version: Column[Optional[str]] = Column(
        String(length=255), doc="The output of hh_server --version"
    )

    branch: Column[Optional[str]] = Column(
        String(length=255),
        doc="Branch the commit is based on",
        nullable=True,
        index=True,
    )

    issue_instances = relationship(
        "IssueInstance",
        primaryjoin="Run.id == foreign(IssueInstance.run_id)",
        backref="run",
    )

    status: Column[str] = Column(
        Enum(RunStatus), server_default="finished", nullable=False, index=True
    )

    status_description: Column[Optional[str]] = Column(
        String(length=255), doc="The reason why a run didn't finish", nullable=True
    )

    kind: Column[Optional[str]] = Column(
        String(length=255),
        doc=(
            "Specify different kinds of runs, e.g. MASTER vs. TEST., GKFORXXX, etc. "
            "in the same DB"
        ),
        nullable=True,
        index=True,
    )

    repository: Column[Optional[str]] = Column(
        String(length=255),
        doc=("The repository that static analysis was run on."),
        nullable=True,
    )

    db_version: Column[int] = Column(
        Integer,
        doc="Tracks under which DB version this was written (for migrations)",
        nullable=False,
        default=CURRENT_DB_VERSION,
        server_default="0",
    )

    purge_status: Column[str] = Column(
        Enum(PurgeStatus),
        server_default="unpurged",
        nullable=False,
        doc="Tracks whether Internal deletion jobs have purged unnecessary issue instances "
        + "and trace frames from this run. Should NOT be set to anything but the default in SAPP code.",
    )

    def get_summary(self, **kwargs) -> RunSummary:
        session = Session.object_session(self)

        return RunSummary(
            commit_hash=self.commit_hash,
            differential_id=self.differential_id,
            id=self.id.resolved(),
            job_id=self.job_id,
            num_new_issues=self._get_num_new_issue_instances(session),
            num_total_issues=self._get_num_total_issues(session),
            alarm_counts=self._get_alarm_counts(session),
        )

    def new_issue_instances(self):
        session = Session.object_session(self)
        return (
            session.query(IssueInstance)
            .filter(IssueInstance.run_id == self.id)
            .filter(IssueInstance.is_new_issue.is_(True))
            .all()
        )

    def _get_num_new_issue_instances(self, session) -> int:
        return (
            session.query(IssueInstance)
            .filter(IssueInstance.run_id == self.id)
            .filter(IssueInstance.is_new_issue.is_(True))
            .count()
        )

    def _get_num_total_issues(self, session) -> int:
        return (
            session.query(IssueInstance).filter(IssueInstance.run_id == self.id).count()
        )

    def _get_alarm_counts(self, session) -> Dict[int, int]:
        return dict(
            session.query(Issue.code, func.count(Issue.code))
            .filter(IssueInstance.run_id == self.id)
            .outerjoin(IssueInstance.issue)
            .group_by(Issue.code)
            .all()
        )


class MetaRun(Base):
    """An identifier that represents multiple runs which should be grouped semantically.

    Meta-runs and runs have a many-to-many relationship, and the purpose of a meta-run
    is to allow querying & displaying results for all related runs without having to
    browse each of them separately."""

    __tablename__ = "metaruns"

    id = Column(BIGDBIDType, primary_key=True, autoincrement=False)

    # This is the moral equivalent of job_id, but named in a more intuitive manner.
    # Allows determining the latest meta run for each custom run separately.
    custom_run_name: Column[Optional[str]] = Column(String(length=255), nullable=True)

    date: Column[datetime] = Column(
        DateTime, doc="The date/time the meta-run was generated", nullable=False
    )

    # We want to be able to filter meta-runs by completion. Towards that end, we plan on
    # using the information of number of total runs vs. the number of runs written in
    # the database.
    expected_run_count: Column[Optional[int]] = Column(Integer, nullable=True)

    kind: Column[Optional[str]] = Column(
        String(length=255),
        doc=(
            "Specify different kinds of runs, e.g. MASTER vs. TEST., GKFORXXX, etc. "
            "in the same DB"
        ),
        nullable=True,
        index=True,
    )

    db_version: Column[int] = Column(
        Integer,
        doc="Tracks under which DB version this was written (for migrations)",
        nullable=False,
        default=CURRENT_DB_VERSION,
    )

    status: Column[str] = Column(
        Enum(RunStatus), server_default="finished", nullable=False, index=True
    )


class RunSummary:
    def __init__(
        self,
        commit_hash: Optional[str],
        differential_id: Optional[int],
        id: Optional[int],
        job_id: Optional[str],
        num_new_issues: int,
        num_total_issues: int,
        num_missing_preconditions: Optional[int] = None,
        num_missing_postconditions: Optional[int] = None,
        alarm_counts: Optional[Dict[int, int]] = None,
    ) -> None:
        self.commit_hash = commit_hash
        self.differential_id = differential_id
        self.id = id
        self.job_id = job_id
        self.num_new_issues = num_new_issues
        self.num_total_issues = num_total_issues
        self.num_missing_preconditions = num_missing_preconditions
        self.num_missing_postconditions = num_missing_postconditions
        self.alarm_counts: Dict[int, int] = alarm_counts or {}

    def todict(self) -> Dict[str, Any]:
        return self.__dict__

    @classmethod
    def fromdict(cls, d) -> "RunSummary":
        return cls(**d)


class MetaRunToRunAssoc(Base, PrepareMixin, RecordMixin):
    """The responsibility of filling out the meta-run to run assoc is on the child jobs
    of a larger run.
    """

    __tablename__ = "metarun_run_assoc"

    meta_run_id = Column(BIGDBIDType, nullable=False, primary_key=True)
    run_id = Column(BIGDBIDType, nullable=False, primary_key=True)
    meta_run = relationship(
        "MetaRun",
        primaryjoin=("MetaRunToRunAssoc.meta_run_id == foreign(MetaRun.id)"),
        uselist=False,
    )
    run = relationship(
        "Run",
        primaryjoin=("MetaRunToRunAssoc.run_id == foreign(Run.id)"),
        uselist=False,
    )

    run_label = Column(
        String(length=1024),
        nullable=True,
        doc="Optional label associated with a child run (eg. Buck target)",
    )

    @classmethod
    def merge(cls, session, items):
        return cls._merge_assocs(session, items, cls.meta_run_id, cls.run_id)


class TraceFrameLeafAssoc(Base, PrepareMixin, RecordMixin):

    __tablename__ = "trace_frame_message_assoc"

    trace_frame_id = Column(BIGDBIDType, nullable=False, primary_key=True)

    leaf_id = Column("message_id", BIGDBIDType, nullable=False, primary_key=True)

    # The minimum trace length unfortunately can be off and actually lead to
    # loops. This is a known problem and any code generating traces should
    # additionally have cycle detection.
    trace_length: Column[Optional[int]] = Column(
        Integer, doc="minimum trace length to the given leaf", nullable=True
    )

    trace_frame = relationship(
        "TraceFrame",
        primaryjoin=("TraceFrameLeafAssoc.trace_frame_id == foreign(TraceFrame.id)"),
        uselist=False,
    )

    leaves = relationship(
        "SharedText",
        primaryjoin="TraceFrameLeafAssoc.leaf_id == foreign(SharedText.id)",
        uselist=False,
    )

    @classmethod
    def merge(cls, session, items):
        return cls._merge_assocs(session, items, cls.trace_frame_id, cls.leaf_id)


class IssueInstanceFixInfo(Base, PrepareMixin, RecordMixin):
    __tablename__ = "issue_instance_fix_info"

    # pyre-fixme[8]: Attribute has type `DBID`; used as `Column[typing.Any]`.
    id: DBID = Column(BIGDBIDType, nullable=False, primary_key=True)

    fix_info: Column[str] = Column(
        String(length=INNODB_MAX_INDEX_LENGTH), nullable=False
    )

    issue_instance = relationship(
        "IssueInstance",
        primaryjoin=("foreign(IssueInstance.fix_info_id) == IssueInstanceFixInfo.id"),
        uselist=False,
    )


class TraceFrame(Base, PrepareMixin, RecordMixin):

    __tablename__ = "trace_frames"

    __table_args__ = (
        Index("ix_traceframe_run_caller_port", "run_id", "caller_id", "caller_port"),
        Index("ix_traceframe_run_callee_port", "run_id", "callee_id", "callee_port"),
    )

    # pyre-fixme[8]: Attribute has type `DBID`; used as `Column[typing.Any]`.
    id: DBID = Column(BIGDBIDType, nullable=False, primary_key=True)

    kind: Column[str] = Column(Enum(TraceKind), nullable=False, index=False)

    caller_id = Column(BIGDBIDType, nullable=False, server_default="0", default=0)

    caller = relationship(
        "SharedText",
        primaryjoin="foreign(SharedText.id) == TraceFrame.caller_id",
        uselist=False,
    )

    # pyre-fixme[8]: Attribute has type `str`; used as `Column[str]`.
    caller_port: str = Column(
        String(length=INNODB_MAX_INDEX_LENGTH),
        nullable=False,
        server_default="",
        doc="The caller port of this call edge",
    )

    callee_id = Column(BIGDBIDType, nullable=False, server_default="0", default=0)

    callee = relationship(
        "SharedText",
        primaryjoin="foreign(SharedText.id) == TraceFrame.callee_id",
        uselist=False,
    )

    callee_location = Column(
        SourceLocationType,
        nullable=False,
        doc="The location of the callee in the source code (line|start|end)",
    )

    # pyre-fixme[8]: Attribute has type `str`; used as `Column[str]`.
    callee_port: str = Column(
        String(length=INNODB_MAX_INDEX_LENGTH),
        nullable=False,
        server_default="",
        doc="The callee port of this call edge'",
    )

    filename_id = Column(BIGDBIDType, nullable=False, server_default="0", default=0)

    run_id = Column("run_id", BIGDBIDType, nullable=False, index=False)

    type_interval_lower: Column[Optional[int]] = Column(
        Integer, nullable=True, doc="Class interval lower-bound (inclusive)"
    )

    type_interval_upper: Column[Optional[int]] = Column(
        Integer, nullable=True, doc="Class interval upper-bound (inclusive)"
    )

    preserves_type_context: Column[bool] = Column(
        Boolean,
        default=False,
        server_default="0",
        nullable=False,
        doc="Whether the call preserves the calling type context",
    )

    titos = Column(
        SourceLocationsType,
        doc="Locations of TITOs aka abductions for the trace frame",
        nullable=False,
        server_default="",
    )

    reachability: Column[str] = Column(
        Enum(FrameReachability),
        server_default="unreachable",
        nullable=False,
        doc="Reachability of this trace frame, for deletion purposes. "
        + "Is set by internal jobs and should NOT be set to anything but the default in SAPP code.",
    )

    annotations = relationship(
        "TraceFrameAnnotation",
        primaryjoin=("TraceFrame.id == foreign(TraceFrameAnnotation.trace_frame_id)"),
        uselist=True,
    )

    leaves = association_proxy("leaf_assoc", "leaves")

    leaf_assoc = relationship(
        "TraceFrameLeafAssoc",
        primaryjoin=("TraceFrame.id == foreign(TraceFrameLeafAssoc.trace_frame_id)"),
        uselist=True,
    )

    issue_instances = association_proxy("trace_frame_issue_instance", "issue_instance")

    trace_frame_issue_instance = relationship(
        "IssueInstanceTraceFrameAssoc",
        primaryjoin=(
            "TraceFrame.id == foreign(IssueInstanceTraceFrameAssoc.trace_frame_id)"
        ),
    )

    leaf_mapping: Set[LeafMapping] = set()

    @staticmethod
    def type_intervals_match_or_ignored(
        caller_start: Optional[int],
        caller_end: Optional[int],
        caller_preserves: Optional[bool],
        callee_start: Optional[int],
        callee_end: Optional[int],
        callee_preserves: Optional[bool],
    ) -> bool:
        """
        returns whether or not to filter based on comparing the type intervals between
        the "caller" trace_frame and the "callee" trace_frame.
        This works both backwards and forwards
        """
        if (
            caller_start is None
            or caller_end is None
            or callee_start is None
            or callee_end is None
            or not callee_preserves
        ):
            # in this case we cannot filter out frames
            return True

        assert caller_start <= caller_end
        assert callee_start <= callee_end

        if caller_start <= callee_start and callee_end <= caller_end:
            # we have a match so we don't filter out the frame
            # in other words for this callee frame the callee is a subset
            # (or the same type) of the callee
            return True

        # we can filter out and we don't have a match return
        # no-match
        # Note that this can happen in a 2 cases
        # (In both cases the caller and callee frams are part of the same base type
        # since we know that the callee 'preserves type')
        # 1. the caller is subset of the callee frame.
        # (e.g. we know the caller is Dog and and the callee could have a trace-frame that
        # allows any animal to traverse.)
        # 2. the caller is an adjacent type
        # (e.g. we know the caller is a Dog and the callee could have a Cat option
        # that needs to be filtered.)
        return False


# Extra bits of information we can show on a TraceFrame.
# This may be a message description, or it may be the start of another series
# of traces leading to some other leaf. TraceFrameAnnotationTraceFrameAssoc
# contains the first hop towards that leaf..
class TraceFrameAnnotation(Base, PrepareMixin, RecordMixin):

    __tablename__ = "trace_frame_annotations"

    # pyre-fixme[8]: Attribute has type `DBID`; used as `Column[typing.Any]`.
    id: DBID = Column(BIGDBIDType, nullable=False, primary_key=True)

    location = Column(
        SourceLocationType, nullable=False, doc="The location for the message"
    )

    kind: Column[Optional[str]] = Column(String(length=255), nullable=True, index=True)

    # pyre-fixme[8]: Attribute has type `str`; used as `Column[str]`.
    message: str = Column(
        String(length=4096),
        doc="Message describing info about the trace",
        nullable=False,
    )

    leaf_id = Column(BIGDBIDType, nullable=True)
    leaf = relationship(
        "SharedText",
        primaryjoin="foreign(SharedText.id) == TraceFrameAnnotation.leaf_id",
        uselist=False,
    )

    # pyre-fixme[8]: Attribute has type `Optional[str]`; used as `Column[str]`.
    link: Optional[str] = Column(
        String(length=4096),
        doc="An optional URL linking the message to more info (Quandary)",
        nullable=True,
    )

    # pyre-fixme[8]: Attribute has type `Optional[str]`; used as `Column[str]`.
    trace_key: Optional[str] = Column(
        String(length=INNODB_MAX_INDEX_LENGTH),
        nullable=True,
        doc="Link to possible pre/post traces (caller_condition).",
    )

    # pyre-fixme[8]: Attribute has type `DBID`; used as `Column[typing.Any]`.
    trace_frame_id: DBID = Column(BIGDBIDType, nullable=False, index=True)
    trace_frame = relationship(
        "TraceFrame",
        primaryjoin=("TraceFrame.id == foreign(TraceFrameAnnotation.trace_frame_id)"),
        uselist=True,
    )

    child_trace_frames = association_proxy(
        "trace_frame_annotation_trace_frame", "trace_frame"
    )
    trace_frame_annotation_trace_frame = relationship(
        "TraceFrameAnnotationTraceFrameAssoc",
        primaryjoin=(
            "TraceFrameAnnotation.id == "
            "foreign(TraceFrameAnnotationTraceFrameAssoc.trace_frame_annotation_id)"
        ),
    )


# A TraceFrameAnnotation may indicate more traces branching out from a trace
# frame towards a different leaf/trace kind. In that case, this assoc describes
# the first hop trace frame from the annotation. It is similar to
# IssueInstanceTraceFrameAssoc, which indicates the first hop trace frame from
# the issue instance.
class TraceFrameAnnotationTraceFrameAssoc(Base, PrepareMixin, RecordMixin):

    __tablename__ = "trace_frame_annotation_trace_frame_assoc"

    trace_frame_annotation_id = Column(
        "trace_frame_annotation_id", BIGDBIDType, primary_key=True, nullable=False
    )

    trace_frame_id = Column(
        "trace_frame_id", BIGDBIDType, primary_key=True, nullable=False, index=True
    )

    trace_frame_annotation = relationship(
        "TraceFrameAnnotation",
        primaryjoin=(
            "TraceFrameAnnotationTraceFrameAssoc.trace_frame_annotation_id == "
            "foreign(TraceFrameAnnotation.id)"
        ),
        uselist=False,
    )

    trace_frame = relationship(
        "TraceFrame",
        primaryjoin=(
            "TraceFrameAnnotationTraceFrameAssoc.trace_frame_id == "
            "foreign(TraceFrame.id)"
        ),
        uselist=False,
    )

    @classmethod
    def merge(cls, session, items):
        return cls._merge_assocs(
            session, items, cls.trace_frame_annotation_id, cls.trace_frame_id
        )


class WarningMessage(Base):
    __tablename__ = "warning_messages"

    code: Column[int] = Column(Integer, autoincrement=False, primary_key=True)

    message: Column[str] = Column(String(length=4096), nullable=False)


class WarningCodeCategory(enum.Enum):
    # Do NOT reorder the enums. Depending on the type of database, existing
    # DBs may have these enums represented internally as ints based on the
    # order shown here, and changing it here messes up existing data. This
    # also means that new enums should be added AT THE END of the list.
    bug = enum.auto()
    code_smell = enum.auto()

    @classproperty
    def BUG(cls):  # noqa
        return cls.bug

    @classproperty
    def CODE_SMELL(cls):  # noqa
        return cls.code_smell


class WarningCodeProperties(Base):
    """Contains properties describing each warning code"""

    __tablename__ = "warning_code_properties"

    code: Column[int] = Column(
        Integer,
        autoincrement=False,
        nullable=False,
        primary_key=True,
        doc="Code identifiying the issue type",
    )

    category: Column[Optional[str]] = Column(
        Enum(WarningCodeCategory),
        nullable=True,
        index=False,
        # pyre-fixme[6]: Expected `str` for 4th param but got `Tuple[str]`.
        doc=(
            "The category of problems that issues in with this warning code "
            "can result in ",
        ),
    )

    new_issue_rate: Column[Optional[Decimal]] = Column(
        Float,
        nullable=True,
        index=False,
        doc="Average number of new issues per day (computed column)",
    )

    bug_count: Column[Optional[int]] = Column(
        Integer,
        nullable=True,
        index=False,
        doc="Number of issues in this category (computed column)",
    )

    avg_trace_len: Column[Optional[Decimal]] = Column(
        Float, nullable=True, index=False, doc="Deprecated. See avg_fwd/bwd_trace_len"
    )

    avg_fwd_trace_len: Column[Optional[Decimal]] = Column(
        Float,
        nullable=True,
        index=False,
        # pyre-fixme[6]: Expected `str` for 4th param but got `Tuple[str]`.
        doc=(
            "Average (min) length of forward traces for the given warning code "
            "(computed column)",
        ),
    )

    avg_bwd_trace_len: Column[Optional[Decimal]] = Column(
        Float,
        nullable=True,
        index=False,
        # pyre-fixme[6]: Expected `str` for 4th param but got `Tuple[str]`.
        doc=(
            "Average (min) length of backward traces for the given warning "
            "code (computed column)",
        ),
    )

    snr: Column[Optional[Decimal]] = Column(
        Float,
        nullable=True,
        index=False,
        doc=(
            "Signal to noise ratio based on triaged issues (computed column). "
            "Ratio of (valid + bad practice) to (false positive + don't care)"
        ),
    )

    is_snr_significant: Column[Optional[bool]] = Column(
        Boolean,
        nullable=True,
        index=False,
        doc=(
            "True if we are confident about the snr (computed column). "
            "Depends on percentage of triaged issues and number of issues."
        ),
    )

    discoverable: Column[Optional[bool]] = Column(
        Boolean,
        nullable=True,
        index=False,
        doc="True if an attacker can discover the issue",
    )

    health_score: Column[Optional[Decimal]] = Column(
        Float,
        nullable=True,
        index=False,
        doc=(
            "Scoring for the health of the warning code, between 0 and 1, "
            "based on the values in the other columns (computed column)"
        ),
    )

    notes: Column[Optional[str]] = Column(
        String(length=4096),
        nullable=True,
        index=False,
        doc="Free form field for note-taking",
    )


class RunOrigin(Base, PrepareMixin, RecordMixin):
    """This table associates runs with metadata concerning how the run was built, which we
    call run origins. An example of run origins is Buck targets."""

    __tablename__ = "run_origins"

    id = Column(BIGDBIDType, nullable=False, primary_key=True)
    run_id = Column(BIGDBIDType, nullable=False, index=True)
    origin = Column(String(length=255), nullable=False)

    run = relationship(
        "Run",
        primaryjoin=("RunOrigin.run_id == foreign(Run.id)"),
        uselist=False,
    )

    @classmethod
    def merge(cls, session, items):
        return cls._merge_by_key(session, items, cls.run_id)


class PrimaryKey(Base, PrimaryKeyBase):
    pass


class PrimaryKeyGenerator(PrimaryKeyGeneratorBase):

    PRIMARY_KEY: Type = PrimaryKey

    QUERY_CLASSES: Set[Type] = {
        Issue,
        IssueInstance,
        IssueInstanceFixInfo,
        SharedText,
        Run,
        TraceFrame,
        TraceFrameAnnotation,
    }


def create(db: DB) -> None:
    try:
        Base.metadata.create_all(db.engine)
    except NoSuchTableError:
        pass


convert_sqlalchemy_type.register(SourceLocationType)(convert_column_to_string)
convert_sqlalchemy_type.register(BIGDBIDType)(convert_column_to_int_or_id)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import json
from operator import itemgetter
from pathlib import Path
from typing import List

import click
from sqlalchemy.orm import aliased
from sqlalchemy.orm.util import AliasedClass

from .cli_lib import require_option
from .models import Issue, IssueInstance, SharedText, TraceFrame

MessageText: AliasedClass = aliased(SharedText)
FilenameText: AliasedClass = aliased(SharedText)
CallerText: AliasedClass = aliased(SharedText)
CalleeText: AliasedClass = aliased(SharedText)


@click.command()
@click.pass_context
@click.option("--run-id", type=int, required=True)
@click.argument(
    "filenames",
    type=click.Path(exists=True, readable=True, resolve_path=True),
    nargs=-1,
    required=True,
)
def lint(click_ctx: click.Context, run_id: int, filenames: List[str]) -> None:
    """Output DB models in a lint-friendly format"""
    ctx = click_ctx.obj
    require_option(click_ctx, "repository")

    paths = [Path(p).resolve() for p in filenames]
    root = Path(ctx.repository).resolve()
    relative = [str(Path(f).relative_to(root)) for f in paths]

    with ctx.database.make_session() as session:
        instances = (
            session.query(
                IssueInstance.location,
                FilenameText.contents.label("filename"),
                MessageText.contents.label("message"),
                Issue.code,
            )
            .filter(IssueInstance.run_id == run_id)
            .join(Issue, Issue.id == IssueInstance.issue_id)
            .join(FilenameText, FilenameText.id == IssueInstance.filename_id)
            .filter(FilenameText.contents.in_(relative))
            .join(MessageText, MessageText.id == IssueInstance.message_id)
            .all()
        )

    with ctx.database.make_session() as session:
        frames = (
            session.query(
                TraceFrame.callee_location,
                TraceFrame.kind,
                TraceFrame.callee_port,
                TraceFrame.caller_port,
                CallerText.contents.label("caller"),
                CalleeText.contents.label("callee"),
                FilenameText.contents.label("filename"),
            )
            .filter(TraceFrame.run_id == run_id)
            .join(FilenameText, FilenameText.id == TraceFrame.filename_id)
            .filter(FilenameText.contents.in_(relative))
            .join(CallerText, CallerText.id == TraceFrame.caller_id)
            .join(CalleeText, CalleeText.id == TraceFrame.callee_id)
            .all()
        )

    # pyre-fixme[53]: Captured variable `root` is not annotated.
    # pyre-fixme[3]: Return type must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    def entry(filename, code, message, location):
        return {
            "filename": str(root / filename),
            "code": code,
            "message": message,
            "line": location.line_no,
            "col": location.begin_column,
            "length": location.begin_column + location.end_column + 1,
        }

    lints = [
        entry(i.filename, str(i.code), i.message, i.location) for i in instances
    ] + [
        entry(
            i.filename,
            i.kind.name,
            f"{i.caller}:{i.caller_port} -> {i.callee}->{i.callee_port}",
            i.callee_location,
        )
        for i in frames
    ]

    for lint in sorted(lints, key=itemgetter("filename", "line", "code", "col")):
        click.echo(json.dumps(lint))



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import json
import os
from glob import glob
from pathlib import Path
from typing import IO, Any, Dict, Iterable, List, NamedTuple, Optional, Tuple

from .sharded_files import ShardedFile

METADATA_GLOB = "*metadata.json"


# pyre-fixme[2]: Parameter annotation cannot contain `Any`.
class Metadata(NamedTuple):
    analysis_root: str
    # Used to relativize paths in the results
    repo_roots: List[str] = []
    repository_name: Optional[str] = None
    tool: Optional[str] = None
    analysis_tool_version: Optional[str] = None
    commit_hash: Optional[str] = None
    job_instance: Optional[int] = None
    project: Optional[str] = None
    # Mapping from code to rule metadata.
    # pyre-ignore: we don't have a shape for rules yet.
    rules: Dict[int, Any] = {}
    type_intervals: Dict[Tuple[int, int], str] = {}

    def merge(self, o: "Metadata") -> "Metadata":
        return Metadata(
            analysis_root=self.analysis_root,
            repo_roots=self.repo_roots + o.repo_roots,
            repository_name=self.repository_name or o.repository_name,
            tool=self.tool or o.tool,
            analysis_tool_version=self.analysis_tool_version or o.analysis_tool_version,
            commit_hash=self.commit_hash or o.commit_hash,
            job_instance=self.job_instance or o.job_instance,
            project=self.project or o.project,
            rules={**self.rules, **o.rules},
            type_intervals={**self.type_intervals, **o.type_intervals},
        )


class AnalysisOutputError(Exception):
    pass


class AnalysisOutput(object):
    """Represents analysis output.

    Possible ways to define, in order of high to low precedence:
    - file_handle: Direct IO object for a single file, useful for testing.
    - filename_specs: List of file names or sharded file patterns.
    - filename_glob and directory: All the files matching the glob in the given directory.
      Avoid patterns like '*', which will include extra, non-analysis files
      in the directory such as the metadata.json.

    Access to the output is provided via generators that provide file handles
    to the diagnostics json (issues), or the summary json (pre and post).
    """

    def __init__(
        self,
        *,
        directory: Optional[str] = None,
        filename_specs: Optional[List[str]] = None,
        filename_glob: Optional[str] = None,
        file_handle: Optional[IO[str]] = None,
        metadata: Optional[Metadata] = None,
        tool: Optional[str] = None,
    ) -> None:
        self.directory = directory
        self.filename_specs: List[str] = filename_specs or []
        self.filename_glob = filename_glob
        self.file_handle = file_handle
        self.metadata = metadata
        self.tool = tool

        if filename_specs is [] and file_handle and hasattr(file_handle, "name"):
            self.filename_specs = [file_handle.name]

    def __str__(self) -> str:
        if self.directory:
            return f"AnalysisOutput({repr(self.directory)})"

        return f"AnalysisOutput({repr(self.filename_specs)})"

    @classmethod
    def from_strs(cls, identifiers: List[str]) -> "AnalysisOutput":
        if len(identifiers) > 1:
            return cls.from_directories(identifiers)
        else:
            return cls.from_str(identifiers[0])

    @classmethod
    def from_str(cls, identifier: str) -> "AnalysisOutput":
        if os.path.isdir(identifier):
            return cls.from_directory(identifier)
        elif os.path.isfile(identifier):
            return cls.from_file(identifier)
        elif os.path.isdir(os.path.dirname(identifier)) and "@" in os.path.basename(
            identifier
        ):
            return cls.from_file(identifier)
        else:
            raise AnalysisOutputError(f"Unrecognized identifier `{identifier}`")

    @classmethod
    def from_directories(cls, directories: List[str]) -> "AnalysisOutput":
        """
        Aggregates several analysis output directories (each of which may themselves be sharded)
        into one AnalysisOutput object. Used for distributed runs of Zoncolan.

        Only supports `filename_spec` in the metadata.json to declare analysis output;
        `filename_glob` and legacy `filenames` are not supported.

        Metadata is naively merged.
        """

        main_metadata = None
        filename_specs = []

        for directory in directories:
            if not os.path.isdir(directory):
                raise AnalysisOutputError(f"`{directory}` is not a directory")
            metadata = {}
            for file in glob(os.path.join(directory, METADATA_GLOB)):
                with open(file) as f:
                    metadata.update(json.load(f))

            if "filename_spec" in metadata:
                filename_specs.append(
                    os.path.join(directory, os.path.basename(metadata["filename_spec"]))
                )

            repo_root = metadata.get("repo_root")
            analysis_root = metadata["root"]
            rules = {rule["code"]: rule for rule in metadata.get("rules", [])}
            this_metadata = Metadata(
                analysis_tool_version=metadata["version"],
                commit_hash=metadata.get("commit"),
                analysis_root=analysis_root,
                repo_roots=[repo_root],
                job_instance=metadata.get("job_instance"),
                tool=metadata.get("tool"),
                repository_name=metadata.get("repository_name"),
                project=metadata.get("project"),
                rules=rules,
                type_intervals=cls._get_interval_dict(metadata),
            )
            if not main_metadata:
                main_metadata = this_metadata
            else:
                main_metadata = main_metadata.merge(this_metadata)
        return cls(
            filename_specs=filename_specs,
            metadata=main_metadata,
        )

    @classmethod
    def from_directory(cls, directory: str) -> "AnalysisOutput":
        metadata = {}
        for file in glob(os.path.join(directory, METADATA_GLOB)):
            with open(file) as f:
                metadata.update(json.load(f))

        filename_specs = []
        filename_glob = None
        if "filename_spec" in metadata:
            filename_specs = [
                os.path.join(directory, os.path.basename(metadata["filename_spec"]))
            ]
        elif "filename_glob" in metadata:
            filename_glob = metadata["filename_glob"]
            if not filename_glob:
                raise AnalysisOutputError(
                    f"Empty 'filename_glob' not allowed. In {METADATA_GLOB}, "
                    "Use either 'filename_spec' or specify something in "
                    "'filename_glob'."
                )
        else:
            # Legacy
            filename_specs = [
                os.path.join(directory, os.path.basename(metadata["filenames"][0]))
            ]

        repo_root = metadata.get("repo_root")
        analysis_root = metadata["root"]

        rules = {rule["code"]: rule for rule in metadata.get("rules", [])}

        return cls(
            directory=directory,
            filename_specs=filename_specs,
            filename_glob=filename_glob,
            metadata=Metadata(
                analysis_tool_version=metadata["version"],
                commit_hash=metadata.get("commit"),
                analysis_root=analysis_root,
                repo_roots=[repo_root],
                job_instance=metadata.get("job_instance"),
                tool=metadata.get("tool"),
                repository_name=metadata.get("repository_name"),
                project=metadata.get("project"),
                rules=rules,
                type_intervals=cls._get_interval_dict(metadata),
            ),
        )

    @classmethod
    def from_file(cls, file_name: str) -> "AnalysisOutput":
        # """Pass in either a single file name or a sharded file pattern.
        # Performs early validation by 1) opening the file if it is a single file,
        # or 2) computing and checking the file shards.
        # """
        return cls(filename_specs=[file_name])

    @classmethod
    def from_handle(cls, file_handle: IO[str]) -> "AnalysisOutput":
        return cls(file_handle=file_handle)

    def file_handles(self) -> Iterable[IO[str]]:
        """Generates all file handles represented by the analysis.
        This method manages closing the handles, no cleanup by the caller is needed.
        """
        if self.file_handle:
            yield self.file_handle
            self.file_handle.close()
            self.file_handle = None
        else:
            for name in self.file_names():
                with open(name, "r") as f:
                    yield f

    def file_names(self) -> Iterable[str]:
        """Generates all file names that are used to generate file_handles."""
        filename_specs = self.filename_specs
        filename_glob = self.filename_glob
        for spec in filename_specs:
            if self._is_sharded(spec):
                yield from ShardedFile(spec).get_filenames()
            else:
                yield spec

        if filename_glob is not None:
            directory = self.directory
            assert directory is not None
            # str() cast to convert the returned Path to string for a
            # consistent return type.
            for path in Path(directory).glob(filename_glob):
                yield str(path)

    @classmethod
    def _is_sharded(cls, spec: str) -> bool:
        return "@" in spec

    def has_sharded(self) -> bool:
        return any(self._is_sharded(spec) for spec in self.filename_specs)

    @classmethod
    def _get_interval_dict(cls, metadata: Dict[str, Any]) -> Dict[Tuple[int, int], str]:
        ret_dict = {}
        for entry in metadata.get("intervals", []):
            interval = entry["interval"]
            if not interval:
                continue
            ret_dict[(interval["start"], interval["finish"])] = entry["type"]
        return ret_dict



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import logging
from typing import Tuple

from ..db import DB
from ..models import create as create_models
from . import DictEntries, PipelineStep, Summary

log: logging.Logger = logging.getLogger("sapp")


class CreateDatabase(PipelineStep[DictEntries, DictEntries]):
    def __init__(self, database: DB) -> None:
        super().__init__()
        self.database = database

    def run(self, input: DictEntries, summary: Summary) -> Tuple[DictEntries, Summary]:
        create_models(self.database)
        return input, summary



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""Bulk saving objects for performance
"""

import logging
from typing import Any, Dict, Optional

from .db import DB
from .decorators import log_time
from .iterutil import split_every
from .models import (
    Issue,
    IssueInstance,
    IssueInstanceFixInfo,
    IssueInstanceSharedTextAssoc,
    IssueInstanceTraceFrameAssoc,
    PrimaryKeyGenerator,
    SharedText,
    TraceFrame,
    TraceFrameAnnotation,
    TraceFrameAnnotationTraceFrameAssoc,
    TraceFrameLeafAssoc,
)

# pyre-fixme[5]: Global expression must be annotated.
log = logging.getLogger("sapp")


class BulkSaver:
    """Stores new objects created within a run and bulk save them"""

    # order is significant, objects will be saved in this order.
    SAVING_CLASSES_ORDER = [
        SharedText,
        Issue,
        IssueInstanceFixInfo,
        IssueInstance,
        IssueInstanceSharedTextAssoc,
        TraceFrame,
        IssueInstanceTraceFrameAssoc,
        TraceFrameAnnotation,
        TraceFrameLeafAssoc,
        TraceFrameAnnotationTraceFrameAssoc,
    ]

    BATCH_SIZE = 30000

    # pyre-fixme[3]: Return type must be annotated.
    def __init__(self, primary_key_generator: Optional[PrimaryKeyGenerator] = None):
        # pyre-fixme[4]: Attribute must be annotated.
        self.primary_key_generator = primary_key_generator or PrimaryKeyGenerator()
        self.saving: Dict[str, Any] = {}
        for cls in self.SAVING_CLASSES_ORDER:
            self.saving[cls.__name__] = []

    # pyre-fixme[3]: Return type must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    def add(self, item):
        assert item.model in self.SAVING_CLASSES_ORDER, (
            "%s should be added with session.add()" % item.model.__name__
        )
        self.saving[item.model.__name__].append(item)

    # pyre-fixme[3]: Return type must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    def add_all(self, items):
        if items:
            assert items[0].model in self.SAVING_CLASSES_ORDER, (
                "%s should be added with session.add_all()" % items[0].model.__name__
            )
            self.saving[items[0].model.__name__].extend(items)

    # pyre-fixme[3]: Return type must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    def get_items_to_add(self, cls):
        return self.saving[cls.__name__]

    # pyre-fixme[3]: Return type must be annotated.
    def save_all(self, database: DB):
        saving_classes = [
            cls
            for cls in self.SAVING_CLASSES_ORDER
            if len(self.saving[cls.__name__]) != 0
        ]

        item_counts = {
            cls.__name__: len(self.get_items_to_add(cls)) for cls in saving_classes
        }

        with database.make_session() as session:
            pk_gen = self.primary_key_generator.reserve(
                session, saving_classes, item_counts
            )

        for cls in saving_classes:
            log.info("Saving %s...", cls.__name__)
            self._save(database, cls, pk_gen)

    @log_time
    # pyre-fixme[3]: Return type must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    def _save(self, database: DB, cls, pk_gen: PrimaryKeyGenerator):
        # We sort keys because bulk insert uses executemany, but it can only
        # group together sequential items with the same keys. If we are scattered
        # then it does far more executemany calls, and it kills performance.
        with database.make_session() as session:
            items = sorted(
                cls.prepare(session, pk_gen, consume(self.saving[cls.__name__])),
                key=lambda k: list(k.keys()),
            )

        # bulk_insert_mappings should only be used for new objects.
        # To update an existing object, just modify its attribute(s)
        # and call session.commit()
        for group in split_every(self.BATCH_SIZE, items):
            with database.make_session() as session:
                session.bulk_insert_mappings(cls, group, render_nulls=True)
                session.commit()

    def add_trace_frame_leaf_assoc(
        self, message: SharedText, trace_frame: TraceFrame, depth: Optional[int]
    ) -> None:
        self.add(
            TraceFrameLeafAssoc.Record(
                trace_frame_id=trace_frame.id, leaf_id=message.id, trace_length=depth
            )
        )

    # pyre-fixme[3]: Return type must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    def add_issue_instance_trace_frame_assoc(self, issue_instance, trace_frame):
        self.add(
            IssueInstanceTraceFrameAssoc.Record(
                issue_instance_id=issue_instance.id, trace_frame_id=trace_frame.id
            )
        )

    # pyre-fixme[3]: Return type must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    def add_issue_instance_shared_text_assoc(self, issue_instance, shared_text):
        self.add(
            IssueInstanceSharedTextAssoc.Record(
                issue_instance_id=issue_instance.id, shared_text_id=shared_text.id
            )
        )

    # pyre-fixme[3]: Return type must be annotated.
    def add_trace_frame_annotation_trace_frame_assoc(
        self,
        # pyre-fixme[2]: Parameter must be annotated.
        trace_frame_annotation,
        # pyre-fixme[2]: Parameter must be annotated.
        trace_frame,
    ):
        self.add(
            TraceFrameAnnotationTraceFrameAssoc.Record(
                trace_frame_annotation_id=trace_frame_annotation.id,
                trace_frame_id=trace_frame.id,
            )
        )

    # pyre-fixme[3]: Return type must be annotated.
    def dump_stats(self):
        stat_str = ""
        for cls in self.SAVING_CLASSES_ORDER:
            stat_str += "%s: %d\n" % (cls.__name__, len(self.saving[cls.__name__]))
        return stat_str


# pyre-fixme[3]: Return type must be annotated.
# pyre-fixme[2]: Parameter must be annotated.
def consume(lst):
    while len(lst) > 0:
        yield lst.pop()



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import logging
from collections import Counter
from dataclasses import dataclass
from typing import Dict, Iterable, List, Optional, Set, Tuple, Union

from .models import DBID, SharedTextKind, TraceFrame, TraceFrameAnnotation, TraceKind
from .trace_graph import TraceGraph

log: logging.Logger = logging.getLogger("sapp")


# Union for queue to recompute trace lengths
@dataclass(frozen=True)
class SearchAction:
    frame: TraceFrame
    remaining_length: int
    leaves: Set[int]


@dataclass(frozen=True)
class ComputeMinAction:
    frame: TraceFrame
    leaves: Set[int]


Action = Union[SearchAction, ComputeMinAction]

# frame_id -> leaf_id -> distance
# Represents that we visited the frame_id looking for leaf kind leaf before.
# There are two cases:
#   distance >= 0 -> we found the leaf within distance hops
#   distance < 0 -> we didn't find the leaf within -distance hops
#
# NOTE: storing the "negative" distance caputres that we did search for leaf
# before and couldn't find it within that many hops. This means that a future
# search with a distance remaining that is less or equal to the previously
# failed distance will also fail. We only revisit if the new visit has more
# "remaining" hops left.
#
# IMPORTANT: The leaf kinds are always the normalized transform kinds, i.e. with
# '@' replaced by ':'. This means we have to be careful at actual leaf-frames to
# compare with the normalized kinds, as well as when updating a frame's trace
# lengths.
Visited = Dict[int, Dict[int, int]]


class TrimmedTraceGraph(TraceGraph):
    """Represents a trimmed graph that is constructed from a bigger TraceGraph
    based on issues that have traces involving a set of affected files or
    directories.
    """

    def __init__(
        self, affected_files: List[str], affected_issues_only: bool = False
    ) -> None:
        """Creates an empty TrimmedTraceGraph."""
        super().__init__()
        self._affected_files = affected_files
        self._affected_issues_only = affected_issues_only
        self._visited_trace_frame_ids: Set[int] = set()

    def populate_from_trace_graph(self, graph: TraceGraph) -> None:
        """Populates this graph from the given one based on affected_files"""
        # Track which trace frames have been visited as we populate the full
        # traces of the graph.
        self._visited_trace_frame_ids: Set[int] = set()

        self._populate_affected_issues(graph)

        if not self._affected_issues_only:
            # Finds issues from the conditions and saves them.
            # Also saves traces that have been trimmed to the affected
            # conditions.
            self._populate_issues_from_affected_trace_frames(graph)

            # Traces populated above may be missing all traces because
            # _populate_issues_from_affected_trace_frames only populates
            # traces that reach the affected conditions in one direction. We
            # may need to populate traces in other directions too.
            #
            # For example:
            #
            # Issue_x reaches affected_file_x via postcondition_x (forward
            # trace, i.e. trace leading to source). None of its backward
            # traces (leading to sinks) reach the affected files.
            #
            # _populate_issues_from_affected_trace_frames would have copied its
            # forward traces and trimmed it to those reaching postcondition_x.
            # We cannot blindly populate all forward traces in this case as
            # branches not leading to postcondition_x are unnecessary.
            #
            # However, in this specific example, all backward traces are needed
            # to give a complete picture of which sinks the issue reaches.
            # The following ensures that.
            for instance_id in self._issue_instances.keys():
                first_hop_ids = self._issue_instance_trace_frame_assoc[instance_id]
                fwd_trace_ids = {
                    tf_id
                    for tf_id in first_hop_ids
                    if self._trace_frames[tf_id].kind == TraceKind.POSTCONDITION
                }
                bwd_trace_ids = {
                    tf_id
                    for tf_id in first_hop_ids
                    if self._trace_frames[tf_id].kind == TraceKind.PRECONDITION
                }
                if len(fwd_trace_ids) == 0:
                    self._populate_issue_trace(
                        graph, instance_id, TraceKind.POSTCONDITION
                    )

                if len(bwd_trace_ids) == 0:
                    self._populate_issue_trace(
                        graph, instance_id, TraceKind.PRECONDITION
                    )

        self._recompute_instance_properties(graph)

    # pyre-fixme[3]: Return type must be annotated.
    def _recompute_instance_properties(self, graph: TraceGraph):
        """Some properties of issue instances will be affected after trimming
        such as min trace length to leaves. This should be called after the
        trimming to re-compute these values.
        """
        callables_histo = Counter(
            inst.callable_id.local_id for inst in self._issue_instances.values()
        )

        # frame_id -> leaf_id -> min_trace
        # where min_trace is negative k, if we didn't reach the leaf in k hops
        visited: Visited = {}

        for inst in self._issue_instances.values():
            # log.info(
            #     "recomputing props for %d",
            #     inst.id.local_id,
            # )

            inst.min_trace_length_to_sources = self._get_min_depth_to_sources(
                visited,
                inst.id.local_id,
                inst.min_trace_length_to_sources,
            )
            inst.min_trace_length_to_sinks = self._get_min_depth_to_sinks(
                visited,
                inst.id.local_id,
                inst.min_trace_length_to_sinks,
            )
            inst.callable_count = callables_histo[inst.callable_id.local_id]

    def _get_min_depth_to_sources(
        self, visited: Visited, instance_id: int, prior: Optional[int]
    ) -> Optional[int]:
        """Returns shortest depth to source from the issue instance. Instances
        have a pre-computed min_trace_length_to_source, but this can change
        after traces get trimmed from the graph. This re-computes it and
        returns the min.
        """
        if prior is None:
            return None

        first_hop_tf_ids = {
            tf_id
            for tf_id in self._issue_instance_trace_frame_assoc[instance_id]
            if self.get_trace_frame_from_id(tf_id).kind == TraceKind.POSTCONDITION
        }
        return self._recompute_trace_length_association(
            visited, first_hop_tf_ids, SharedTextKind.source
        )

    def _get_min_depth_to_sinks(
        self, visited: Visited, instance_id: int, prior: Optional[int]
    ) -> Optional[int]:
        """See get_min_depths_to_sources."""
        if prior is None:
            return None

        first_hop_tf_ids = {
            tf_id
            for tf_id in self._issue_instance_trace_frame_assoc[instance_id]
            if self.get_trace_frame_from_id(tf_id).kind == TraceKind.PRECONDITION
        }
        return self._recompute_trace_length_association(
            visited, first_hop_tf_ids, SharedTextKind.sink
        )

    def _map_info(self, v: Dict[int, int]) -> str:
        return ", ".join(
            [f"{self._get_local_text(key)} -> {d}" for key, d in v.items()]
        )

    def _remaining_leaves(
        self,
        remaining_length: int,
        leaves: Set[int],
        visited: Visited,
        frame_id: int,
    ) -> Set[int]:
        """Given a visit to frame_id with remaining_length and looking for the given
        leaves, compute whether we already have visited this frame under some of
        these leaves. Returns the remaining leaves that will need to be visited
        for the children. There are multiple cases:

        1. we have visited the frame_id before and found a distance to the leaf,
        remove the leaf.

        2. we have visited the frame_id before but didn't finish the visit
        because it didn't lead to the leaf before we ran out of trace length (or
        we are recursively visiting it). In that case, the negative number
        stored is the length we looked for. If we have more remaining_length
        left now, we revisit the leaf, otherwise we don't.

        For the leaves we will search in children, we update the visited state
        to indicate that this is on the stack by adding -remaining_length to the
        visited state.

        """
        assert remaining_length > 0
        if frame_id in visited:
            visited_leaves = visited[frame_id]
            # figure out what needs to be visited still
            # log.info("    old visited: %s", self._map_info(visited_leaves))
            visit_leaves = {
                leaf_id: -remaining_length
                for leaf_id in leaves
                if leaf_id not in visited_leaves
                or visited_leaves[leaf_id] < 0
                and -visited_leaves[leaf_id] < remaining_length
            }
            visited[frame_id].update(visit_leaves)
            # log.info("    new visited: %s", self._map_info(visited[frame_id]))
            return set(visit_leaves.keys())
        else:
            # first time. Put remaining trace lengths (pending)
            visited[frame_id] = {leaf_id: -remaining_length for leaf_id in leaves}
            # log.info("    first visit: %s", self._map_info(visited[frame_id]))
            return leaves

    def _get_text(self, id: DBID) -> str:
        return self.get_shared_text_by_local_id(id.local_id).contents

    def _get_local_text(self, id: int) -> str:
        return self.get_shared_text_by_local_id(id).contents

    def _frame_info(self, frame: TraceFrame) -> str:
        return (
            f"frame_id:{frame.id.local_id} "
            f"caller:{self._get_text(frame.caller_id)} "
            f"caller_port:{frame.caller_port} "
            f"callee:{self._get_text(frame.callee_id)} "
            f"callee_port:{frame.callee_port} "
            f"interval:[{frame.type_interval_lower},{frame.type_interval_upper}]"
        )

    def _recompute_trace_length_association(
        self, visited: Visited, initial_frames: Set[int], leaf_kind: SharedTextKind
    ) -> int:

        """Walks the traces starting at the initial frames with the initial
        corresponding kinds to recompute and store the minimum trace length from each
        reachable frame to the corresponding leaf."""

        max_trace_length = 100
        infinite_trace_length = 9999
        stack: List[Action] = [
            SearchAction(
                frame=self.get_trace_frame_from_id(frame_id),
                remaining_length=max_trace_length,
                leaves=self.get_caller_leaf_kinds_of_frame(
                    self.get_trace_frame_from_id(frame_id)
                ),
            )
            for frame_id in initial_frames
        ]

        while len(stack) > 0:
            todo = stack.pop()
            if isinstance(todo, SearchAction):
                # log.info(
                #     "  search %s, remaining %d leaves: %s",
                #     self._frame_info(todo.frame),
                #     todo.remaining_length,
                #     " ".join(
                #         [self._get_local_text(leaf_id) for leaf_id in todo.leaves]
                #     ),
                # )
                frame_id = todo.frame.id.local_id
                leaves = self._remaining_leaves(
                    todo.remaining_length, todo.leaves, visited, frame_id
                )
                if len(leaves) == 0 or todo.remaining_length <= 1:
                    continue

                # log.info(
                #     "    remaining leaves: %s",
                #     " ".join([self._get_local_text(leaf_id) for leaf_id in leaves]),
                # )

                if self.is_leaf_port(todo.frame.callee_port):
                    actual_leaves = {}
                    for leaf_id in self.get_trace_frame_leaf_ids_by_kind(
                        todo.frame, leaf_kind
                    ):
                        leaf = self.get_shared_text_by_local_id(leaf_id)
                        leaf_id = self.get_transform_normalized_kind_id(leaf)
                        actual_leaves[leaf_id] = 0
                    visited[frame_id].update(actual_leaves)
                    # log.info(
                    #     "    leaf result %s",
                    #     self._map_info(actual_leaves),
                    # )
                    continue

                successors = []
                (successor_frames, succ_leaf_kinds) = self._get_successor_frames(
                    self,
                    leaves,
                    todo.frame,
                )
                if len(succ_leaf_kinds) > 0:
                    for next_frame in successor_frames:
                        successors.append(
                            SearchAction(
                                frame=next_frame,
                                remaining_length=todo.remaining_length - 1,
                                leaves=succ_leaf_kinds,
                            )
                        )
                # Note: list append/pop both work from the tail, so we have to
                # append the ComputeMin first before the search on the children.
                stack.append(
                    ComputeMinAction(
                        todo.frame,
                        leaves,
                    )
                )
                stack.extend(successors)

            elif isinstance(todo, ComputeMinAction):
                visit_result = visited[todo.frame.id.local_id]
                # log.info(
                #     "  compute min %s, leaves: %s",
                #     self._frame_info(todo.frame),
                #     " ".join(
                #         [f"{self._get_local_text(leaf_id)}" for leaf_id in todo.leaves]
                #     ),
                # )
                for leaf_id in todo.leaves:
                    # log.info("    looking for %s", self._get_local_text(leaf_id))
                    (successors, succ_leaves) = self._get_successor_frames(
                        self,
                        {leaf_id},
                        todo.frame,
                    )
                    if len(succ_leaves) > 0:
                        # log.info(
                        #     "      succ_leaves %s",
                        #     ", ".join([self._get_local_text(id) for id in succ_leaves]),
                        # )
                        for succ in successors:
                            for succ_leaf_id, length in visited[
                                succ.id.local_id
                            ].items():
                                if succ_leaf_id in succ_leaves:
                                    if length >= 0 and (
                                        length + 1 < visit_result[leaf_id]
                                        or visit_result[leaf_id] < 0
                                    ):
                                        visit_result[leaf_id] = length + 1
                                    elif (
                                        length < 0
                                        and length - 1 > visit_result[leaf_id]
                                    ):
                                        visit_result[leaf_id] = length - 1
                frame_result = {}
                frame_leaves = self.get_trace_frame_leaf_ids_with_depths(todo.frame)
                for frame_leaf_id in frame_leaves:
                    frame_leaf = self.get_shared_text_by_local_id(frame_leaf_id)
                    if (
                        frame_leaf.kind != SharedTextKind.source
                        and frame_leaf.kind != SharedTextKind.sink
                    ):
                        continue
                    normalized_frame_leaf_id = self.get_transform_normalized_kind_id(
                        frame_leaf
                    )
                    if normalized_frame_leaf_id in todo.leaves:
                        if visit_result[normalized_frame_leaf_id] < 0:
                            frame_result[frame_leaf_id] = infinite_trace_length
                        else:
                            frame_result[frame_leaf_id] = visit_result[
                                normalized_frame_leaf_id
                            ]

                # log.info(
                #     "    frame_result: %s",
                #     self._map_info(frame_result),
                # )
                # log.info(
                #     "    visit_result: %s",
                #     self._map_info(visit_result),
                # )
                visited[todo.frame.id.local_id].update(visit_result)
                self.get_trace_frame_leaf_ids_with_depths(todo.frame).update(
                    frame_result
                )

        # compute minimum over all initial frames/leaves
        result = infinite_trace_length
        for frame_id in initial_frames:
            for _, length in visited[frame_id].items():
                if length >= 0 and length < result:
                    result = length
        return result

    def _populate_affected_issues(self, graph: TraceGraph) -> None:
        """Populates the trimmed graph with issues whose locations are in
        affected_files based on data in the input graph. Since these issues
        exist in the affected files, all traces are copied as well.
        """
        affected_instance_ids = [
            instance.id.local_id
            for instance in graph._issue_instances.values()
            if self._is_filename_prefixed_with(
                graph.get_text(instance.filename_id), self._affected_files
            )
        ]

        for instance_id in affected_instance_ids:
            if instance_id in self._issue_instances:
                continue
            self._populate_issue_and_traces(graph, instance_id)

    def _get_sink_kinds(self, graph: TraceGraph, instance_id: int) -> Set[int]:
        kind: SharedTextKind = SharedTextKind.SINK
        sinks = graph.get_issue_instance_shared_texts(instance_id, kind)
        return {sink.id.local_id for sink in sinks}

    def _get_source_kinds(self, graph: TraceGraph, instance_id: int) -> Set[int]:
        kind: SharedTextKind = SharedTextKind.SOURCE
        sources = graph.get_issue_instance_shared_texts(instance_id, kind)
        return {source.id.local_id for source in sources}

    def _get_instance_leaf_ids(self, graph: TraceGraph, instance_id: int) -> Set[int]:
        return self._get_source_kinds(graph, instance_id).union(
            self._get_sink_kinds(graph, instance_id)
        )

    def _populate_issues_from_affected_trace_frames(self, graph: TraceGraph) -> None:
        """TraceFrames found in affected_files should be reachable via some
        issue instance. Follow traces of graph to find them and
        populate this TrimmedGraph with it.
        """

        initial_trace_frames = [
            trace_frame
            for trace_frame in graph._trace_frames.values()
            if self._is_filename_prefixed_with(
                graph.get_text(trace_frame.filename_id), self._affected_files
            )
        ]

        self._populate_issues_from_affected_conditions(
            initial_trace_frames,
            graph,
        )

    def _get_issue_instances_from_frame_id(
        self, graph: TraceGraph, trace_frame_id: int
    ) -> Set[int]:
        return graph._trace_frame_issue_instance_assoc[trace_frame_id]

    def _get_predecessor_frames(
        self, graph: TraceGraph, leaves: Set[int], trace_frame: TraceFrame
    ) -> List[Tuple[TraceFrame, Set[int]]]:
        """Returns predecessor frames paired with leaf kinds to follow for those frames"""
        result = []
        # pyre-fixme[6]: Enums and str are the same but Pyre doesn't think so.
        for trace_frame_id in graph._trace_frames_rev_map[trace_frame.kind][
            (trace_frame.caller_id.local_id, trace_frame.caller_port)
        ]:
            predecessor = graph._trace_frames[trace_frame_id]
            assert predecessor.leaf_mapping is not None
            pred_kinds = graph.compute_prev_leaf_kinds(leaves, predecessor.leaf_mapping)
            result.append((predecessor, pred_kinds))
        return result

    def _get_successor_frames(
        self, graph: TraceGraph, leaves: Set[int], trace_frame: TraceFrame
    ) -> Tuple[List[TraceFrame], Set[int]]:
        """Returns successor frames and successor leaf_kind pair"""
        result = []
        assert trace_frame.leaf_mapping is not None
        succ_kinds = graph.compute_next_leaf_kinds(leaves, trace_frame.leaf_mapping)
        # pyre-fixme[6]: Enums and str are the same but Pyre doesn't think so.
        for trace_frame_id in graph._trace_frames_map[trace_frame.kind][
            (trace_frame.callee_id.local_id, trace_frame.callee_port)
        ]:
            successor = graph._trace_frames[trace_frame_id]
            result.append(successor)
        return (result, succ_kinds)

    def _populate_issues_from_affected_conditions(
        self,
        # pyre-fixme[2]: Parameter must be annotated.
        initial_conditions,
        graph: TraceGraph,
    ) -> None:
        """Helper for populating reachable issue instances from the initial
        pre/postconditions. Also populates conditions/traces reachable from
        these instances. Traces are populated only in the direction this is
        called from: i.e. if initial_conditions are preconditions, only the
        backward trace is populated.

        Params:
        initial_conditions: The initial collection of pre/postconditions to
        start searching for issues from.

        graph: The trace graph to search for issues. Nodes/edges in this graph
        will be copied over to the local state
        """
        visited: Dict[int, Set[int]] = {}
        stack = [
            # We will be using these leaf kinds to look for matching callers. So
            # we need the caller view of the kinds.
            (frame, graph.get_caller_leaf_kinds_of_frame(frame))
            for frame in initial_conditions
        ]

        # Note that parent conditions may not transitively lead to the leaves
        # that its descendents lead to due to special-cased leaf filtering at
        # analysis time. When visiting each condition, we need to track the
        # leaves that we are visiting it from and only visit parent traces that
        # share common leaves along the path.
        while len(stack) > 0:
            condition, leaves = stack.pop()
            cond_id = condition.id.local_id

            if cond_id in visited:
                leaves = leaves - visited[cond_id]
                if len(leaves) == 0:
                    continue
                else:
                    visited[cond_id].update(leaves)
            else:
                visited[cond_id] = leaves

            # Found instance(s) related to the current condition. Yay.
            # This instance may have been found before, but process it again
            # anyway because we need to add the assoc with this condition.
            for instance_id in self._get_issue_instances_from_frame_id(graph, cond_id):
                # Check if the leaves (sources/sinks) of the issue reach
                # the same leaves as the ones relevant to this condition.
                instance = graph._issue_instances[instance_id]
                issue_leaves = set(
                    self._get_instance_leaf_ids(graph, instance.id.local_id)
                )
                common_leaves = issue_leaves.intersection(leaves)
                if len(common_leaves) > 0:
                    if instance_id not in self._issue_instances:
                        self._populate_issue(graph, instance_id)
                    self.add_issue_instance_trace_frame_assoc(instance, condition)

            # Conditions that call this may have originated from other issues,
            # keep searching for parent conditions leading to this one.
            for (next_frame, frame_leaves) in self._get_predecessor_frames(
                graph, leaves, condition
            ):
                if len(frame_leaves) > 0:
                    stack.append((next_frame, frame_leaves))

        # Add traces leading out from initial_conditions, and all visited
        # conditions leading back towards the issues.
        initial_condition_ids = [
            condition.id.local_id for condition in initial_conditions
        ]
        self._populate_trace(graph, initial_condition_ids)
        for frame_id in visited:
            self._add_trace_frame(graph, graph._trace_frames[frame_id])

    def _populate_issue_and_traces(self, graph: TraceGraph, instance_id: int) -> None:
        """Copies an issue over from the given trace graph, including all its
        traces and assocs.
        """
        self._populate_issue(graph, instance_id)
        self._populate_issue_trace(graph, instance_id)

    def _populate_issue_trace(
        self, graph: TraceGraph, instance_id: int, kind: Optional[TraceKind] = None
    ) -> None:
        trace_frame_ids = list(graph._issue_instance_trace_frame_assoc[instance_id])
        instance = graph._issue_instances[instance_id]
        filtered_ids = []
        for trace_frame_id in trace_frame_ids:
            frame = graph._trace_frames[trace_frame_id]
            if kind is None or kind == frame.kind:
                self.add_issue_instance_trace_frame_assoc(instance, frame)
                filtered_ids.append(trace_frame_id)
        self._populate_trace(graph, filtered_ids)

    def _populate_issue(self, graph: TraceGraph, instance_id: int) -> None:
        """Adds an issue to the trace graph along with relevant information
        pertaining to the issue (e.g. instance, fix_info, sources/sinks)
        The issue is identified by its corresponding instance's ID in the input
        trace graph.
        """
        instance = graph._issue_instances[instance_id]
        issue = graph._issues[instance.issue_id.local_id]
        self._populate_shared_text(graph, instance.message_id)
        self._populate_shared_text(graph, instance.filename_id)
        self._populate_shared_text(graph, instance.callable_id)

        self.add_issue_instance(instance)
        self.add_issue(issue)

        if instance_id in graph._issue_instance_fix_info:
            issue_fix_info = graph._issue_instance_fix_info[instance_id]
            self.add_issue_instance_fix_info(instance, issue_fix_info)

        for shared_text_id in graph._issue_instance_shared_text_assoc[instance_id]:
            shared_text = graph._shared_texts[shared_text_id]
            if shared_text_id not in self._shared_texts:
                self.add_shared_text(shared_text)
            self.add_issue_instance_shared_text_assoc(instance, shared_text)

    def _populate_trace(self, graph: TraceGraph, trace_frame_ids: List[int]) -> None:
        """Populates (from the given trace graph) the forward and backward
        traces reachable from the given traces (including input trace frames).
        Make sure to respect trace kind in successors
        """
        while len(trace_frame_ids) > 0:
            trace_frame_id = trace_frame_ids.pop()
            if trace_frame_id in self._visited_trace_frame_ids:
                continue

            trace_frame = graph._trace_frames[trace_frame_id]
            self._add_trace_frame(graph, trace_frame)
            self._visited_trace_frame_ids.add(trace_frame_id)

            trace_frame_ids.extend(
                [
                    next_frame.id.local_id
                    for next_frame in graph.get_next_trace_frames(trace_frame)
                    if next_frame.id.local_id not in self._visited_trace_frame_ids
                ]
            )

    def _add_trace_frame(self, graph: TraceGraph, trace_frame: TraceFrame) -> None:
        """Copies the trace frame from 'graph' to this (self) graph.
        Also copies all the trace_frame-leaf assocs since we don't
        know which ones are needed until we know the issue that reaches it
        """
        trace_frame_id = trace_frame.id.local_id
        self.add_trace_frame(trace_frame)

        annotations = graph.get_condition_annotations(trace_frame_id)
        for annotation in annotations:
            self._add_trace_annotation(graph, annotation)

        self._populate_shared_text(graph, trace_frame.filename_id)
        self._populate_shared_text(graph, trace_frame.caller_id)
        self._populate_shared_text(graph, trace_frame.callee_id)
        for (leaf_id, depth) in graph._trace_frame_leaf_assoc[trace_frame_id].items():
            leaf = graph._shared_texts[leaf_id]
            if leaf_id not in self._shared_texts:
                self.add_shared_text(leaf)
            self.add_trace_frame_leaf_assoc(trace_frame, leaf, depth)

    @staticmethod
    def _is_filename_prefixed_with(filename: str, prefixes: Iterable[str]) -> bool:
        return any(filename.startswith(p) for p in prefixes)

    # pyre-fixme[2]: Parameter must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    def _populate_shared_text(self, graph, id) -> None:
        text = graph._shared_texts[id.local_id]
        if text.id.local_id not in self._shared_texts:
            self.add_shared_text(text)

    def _add_trace_annotation(
        self, graph: TraceGraph, annotation: TraceFrameAnnotation
    ) -> None:
        """Copies the annotation from 'graph' to this (self) graph.
        Also copies children TraceFrames of the annotation (if any). The
        parent TraceFrame of the annotation is NOT copied.
        """
        self.add_trace_annotation(annotation)
        children = graph.get_annotation_trace_frames(annotation.id.local_id)
        child_ids = [child.id.local_id for child in children]
        for child in children:
            self.add_trace_frame_annotation_trace_frame_assoc(annotation, child)
        self._populate_trace(graph, child_ids)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import logging
from collections import defaultdict, deque
from typing import Dict, List, Set, Tuple

from ..models import IssueInstance, SharedTextKind, TraceFrame, TraceKind
from ..trace_graph import TraceGraph
from . import PipelineStep, Summary

log: logging.Logger = logging.getLogger("sapp")


class PerSinkState:
    def __init__(self) -> None:
        self.shared_text_depths: Dict[int, int] = defaultdict(lambda: 999999)


FrameID = int
SinkToSharedTextMap = Dict[int, Set[int]]
SinkToState = Dict[int, PerSinkState]


class PropagateSharedTexts(PipelineStep[TraceGraph, TraceGraph]):  # pyre-fixme[13]
    """For all issues propagate source kinds to all reachable frames leading to
    sinks and propagate features to leaf sinks."""

    def __init__(self, propagate_sources: bool, propagate_features: bool) -> None:
        super().__init__()
        self.summary: Summary
        self.graph: TraceGraph
        self.propagate_sources = propagate_sources
        self.propagate_features = propagate_features
        # pyre-fixme[8]
        self.visited: Dict[FrameID, SinkToState] = defaultdict(
            lambda: defaultdict(lambda: PerSinkState())
        )

    def _subtract_kinds(
        self,
        depth: int,
        kind_map: SinkToSharedTextMap,
        to_remove: SinkToState,
    ) -> SinkToSharedTextMap:
        """Prunes the search space by eliminating sources and features that have already
        been visited at a closer distance than our current depth"""
        result = {}
        for sink_kind, shared_texts in kind_map.items():
            if sink_kind in to_remove:
                shared_text_depths = to_remove[sink_kind].shared_text_depths
                shared_texts = {
                    shared_text
                    for shared_text in shared_texts
                    if shared_text_depths.get(shared_text, depth + 1) > depth
                }
            if len(shared_texts) > 0:
                result[sink_kind] = shared_texts
        return result

    def _update_visited(
        self,
        frame_id: FrameID,
        depth: int,
        kind_map: SinkToSharedTextMap,
    ) -> None:
        for sink_kind, shared_texts in kind_map.items():
            visited_frame = self.visited[frame_id]
            for shared_text in shared_texts:
                my_depth = depth
                # Normally, when we decrease the distance of a source, we want to keep going
                # to decrease the distance on all subsequent frames.
                # But for features, we don't care about this.
                # Just set the depth low immediately when visited regardless of the
                # actual distance visited, so that
                # work is not wasted propagating such decreases.
                kind = self.graph.get_shared_text_by_local_id(shared_text).kind
                if kind is SharedTextKind.FEATURE:
                    my_depth = 0
                visited_frame[sink_kind].shared_text_depths[shared_text] = my_depth

    def _propagate_shared_texts(self, instance: IssueInstance) -> None:
        """Propagate the source kinds and features of this issue instance to all
        reachable sink traces."""
        graph = self.graph

        features = {
            text.id.local_id
            for text in graph.get_issue_instance_shared_texts(
                instance.id.local_id, SharedTextKind.FEATURE
            )
        }
        initial_frames = graph.get_issue_instance_trace_frames(instance)
        initial_source_frames = [
            frame for frame in initial_frames if frame.kind == TraceKind.POSTCONDITION
        ]
        initial_sink_frames = [
            frame for frame in initial_frames if frame.kind == TraceKind.PRECONDITION
        ]
        source_kind_list = [
            graph.get_caller_leaf_kinds_of_frame(frame)
            for frame in initial_source_frames
        ]

        shared_text_kinds = set.union(*source_kind_list).union(features)
        if len(shared_text_kinds) == 0:
            return
        self._propagate_kinds_along_traces(initial_sink_frames, shared_text_kinds)

    def _propagate_kinds_along_traces(
        self, start_frames: List[TraceFrame], new_kinds: Set[int]
    ) -> None:
        graph = self.graph

        queue = deque(
            [
                (
                    start_frame,
                    {
                        sink_id: new_kinds
                        for sink_id in graph.get_caller_leaf_kinds_of_frame(start_frame)
                    },
                    0,
                )
                for start_frame in start_frames
            ]
        )
        while len(queue) > 0:
            frame, kind_map, depth = queue.popleft()
            if len(kind_map) == 0:
                continue

            frame_id = frame.id.local_id
            if frame_id in self.visited:
                kind_map = self._subtract_kinds(depth, kind_map, self.visited[frame_id])
                if len(kind_map) == 0:
                    continue

            self._update_visited(frame_id, depth, kind_map)

            next_frames = self.graph.get_trace_frames_from_caller(
                # pyre-fixme[6]: Expected `TraceKind` for 1st param but got `str`.
                frame.kind,
                frame.callee_id,
                frame.callee_port,
            )

            queue.extend(
                (
                    frame,
                    {
                        leaf_map.callee_leaf: kind_map[leaf_map.caller_leaf]
                        for leaf_map in frame.leaf_mapping
                        if leaf_map.caller_leaf in kind_map
                    },
                    depth + 1,
                )
                for frame in next_frames
            )

    def run(self, input: TraceGraph, summary: Summary) -> Tuple[TraceGraph, Summary]:
        if not (self.propagate_sources or self.propagate_features):
            return input, summary

        graph = input
        self.summary = summary
        self.graph = graph

        if self.propagate_sources:
            log.info("Propagating source kinds to sinks")

        if self.propagate_features:
            log.info("Propagating features to anchor sinks")

        for instance in graph.get_issue_instances():
            self._propagate_shared_texts(instance)

        # Create new assocs based on the visited results
        source_count = 0
        feature_count = 0
        trace_frame_count = 0
        for trace_frame_id, sink_to_state in self.visited.items():
            trace_frame_count += 1
            trace_frame = graph.get_trace_frame_from_id(trace_frame_id)
            is_anchor_port = trace_frame.callee_port.startswith("anchor:")
            for state in sink_to_state.values():
                for shared_text, depth in state.shared_text_depths.items():
                    shared_text_kind = graph.get_shared_text_by_local_id(
                        shared_text
                    ).kind
                    if (
                        self.propagate_sources
                        and shared_text_kind == SharedTextKind.SOURCE
                    ):
                        graph.add_trace_frame_leaf_by_local_id_assoc(
                            trace_frame, shared_text, depth
                        )
                        source_count += 1
                    if (
                        self.propagate_features
                        and is_anchor_port
                        and shared_text_kind == SharedTextKind.FEATURE
                    ):
                        graph.add_trace_frame_leaf_by_local_id_assoc(
                            trace_frame, shared_text, depth=None
                        )
                        feature_count += 1
        log.info(
            f"Added {source_count} source kinds and {feature_count} features to {trace_frame_count} trace frames"
        )

        return graph, summary



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import os
from pathlib import Path
from typing import List, Optional


def find_root(vcs_dirs: List[str]) -> Optional[str]:
    path = Path.cwd().resolve()
    while not path.samefile(path.root):
        if any((path / vcs_dir).exists() for vcs_dir in vcs_dirs):
            return os.fspath(path)
        path = path.parent
    return None



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from __future__ import annotations

import json
import sys
from json import JSONEncoder
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Union

from sqlalchemy import Column, String

from .models import Base

if TYPE_CHECKING:
    from .ui.schema import FeatureCondition, MatchesIsField

if sys.version_info >= (3, 8):
    from typing import TypedDict
else:
    from typing_extensions import TypedDict


class NewMatchesIsAlias(TypedDict):
    operation: str
    value: List[str]


# Previously the field contained values in form of a list, maintain Union for
# backward compatibility.
MatchesIsAlias = Union[NewMatchesIsAlias, List[str]]


class FilterRecord(Base):
    __tablename__ = "filters"

    name: Column[str] = Column(
        String(length=255), nullable=False, unique=True, primary_key=True
    )
    description: Column[Optional[str]] = Column(String(length=1024), nullable=True)

    json: Column[str] = Column(
        String(length=1024), nullable=False, doc="JSON representation of the filter"
    )


class FilterValidationException(Exception):
    pass


class Filter:
    def __init__(self, **kwargs: Any) -> None:
        self.features: List[Dict[str, Union[str, List[str]]]] = kwargs.get(
            "features", []
        )
        self.codes: List[int] = kwargs.get("codes", [])
        self.paths: List[str] = kwargs.get("paths", [])
        self.callables: MatchesIsAlias = kwargs.get("callables", [])
        self.statuses: List[str] = kwargs.get("statuses", [])
        self.source_names: MatchesIsAlias = kwargs.get("source_names", [])
        self.source_kinds: MatchesIsAlias = kwargs.get("source_kinds", [])
        self.sink_names: MatchesIsAlias = kwargs.get("sink_names", [])
        self.sink_kinds: MatchesIsAlias = kwargs.get("sink_kinds", [])
        self.traceLengthFromSources: Optional[List[int]] = kwargs.get(
            "traceLengthFromSources", None
        )
        self.traceLengthToSinks: Optional[List[int]] = kwargs.get(
            "traceLengthToSinks", None
        )
        self.is_new_issue: Optional[bool] = kwargs.get("is_new_issue", None)

        missing_filtering_condition: bool = all(
            getattr(self, key) in (None, []) for key in self._json_filtering_keys()
        )

        if missing_filtering_condition:
            filtering_conditions: List[str] = self._json_filtering_keys()
            raise FilterValidationException(
                f"Error: The filter must have at least one of the following keys: f{filtering_conditions}"
            )

    def _json_filtering_keys(self) -> List[str]:
        return [
            attribute
            for attribute in self.__dict__.keys()
            if not attribute.startswith("__")
            and not callable(attribute)
            and attribute != "name"
            and attribute != "description"
        ]

    def to_json(self) -> str:
        return json.dumps(self, cls=FilterEncoder)

    def format_features_for_query(self) -> List[Tuple[str, List[str]]]:
        formatted_features = []
        if self.features is not None:
            for feature in self.features:
                formatted_features.append((feature["mode"], feature["features"]))
        return formatted_features

    @staticmethod
    def from_query(
        codes: List[int],
        paths: List[str],
        callables: Optional[MatchesIsField],
        statuses: List[str],
        source_names: Optional[MatchesIsField],
        source_kinds: Optional[MatchesIsField],
        sink_names: Optional[MatchesIsField],
        sink_kinds: Optional[MatchesIsField],
        features: Optional[List[FeatureCondition]],
        min_trace_length_to_sinks: Optional[int],
        max_trace_length_to_sinks: Optional[int],
        min_trace_length_to_sources: Optional[int],
        max_trace_length_to_sources: Optional[int],
        is_new_issue: Optional[bool],
    ) -> Filter:

        restructured_features: List[Dict[str, Union[str, List[str]]]] = []
        for filtering_condition in features or []:
            feature_entry = {}
            feature_entry["mode"] = filtering_condition.mode
            feature_entry["features"] = list(filtering_condition.features)
            restructured_features.append(feature_entry)

        traceLengthFromSources: Optional[List[int]] = None
        if (
            min_trace_length_to_sources is not None
            or max_trace_length_to_sources is not None
        ):
            traceLengthFromSources = [
                min_trace_length_to_sources or 0,
                max_trace_length_to_sources or 31,
            ]

        traceLengthToSinks: Optional[List[int]] = None
        if (
            min_trace_length_to_sinks is not None
            or max_trace_length_to_sinks is not None
        ):
            traceLengthToSinks = [
                min_trace_length_to_sinks or 0,
                max_trace_length_to_sinks or 31,
            ]

        return Filter(
            features=restructured_features,
            codes=codes,
            paths=paths,
            callables=callables,
            statuses=statuses,
            source_names=source_names,
            source_kinds=source_kinds,
            sink_names=sink_names,
            sink_kinds=sink_kinds,
            traceLengthFromSources=traceLengthFromSources,
            traceLengthToSinks=traceLengthToSinks,
            is_new_issue=is_new_issue,
        )


class FilterEncoder(JSONEncoder):
    def default(
        self, o: Filter
    ) -> Dict[str, Union[bool, List[int], List[str], Dict[str, Union[str, List[str]]]]]:
        filtering_conditions: Dict[str, Any] = {
            attribute: value
            for attribute, value in o.__dict__.items()
            if value and value != ["%"]
        }
        filtering_conditions.pop("name", None)
        filtering_conditions.pop("description", None)
        return filtering_conditions


class StoredFilter(Filter):
    def __init__(self, name: str, description: str = "", **kwargs: Any) -> None:
        self.name = name
        if self.name is None:
            raise FilterValidationException("Error: A StoredFilter must have a name")
        self.description = description
        if self.description is None:
            raise FilterValidationException(
                "Error: A StoredFilter description cannot be `None`"
            )
        super(StoredFilter, self).__init__(**kwargs)

    @staticmethod
    def from_file(input_path: Path) -> StoredFilter:
        json_blob: Dict[str, Any] = json.loads(input_path.read_text())
        return StoredFilter(**json_blob)

    @staticmethod
    def from_record(record: FilterRecord) -> StoredFilter:
        return StoredFilter(
            record.name, record.description or "", **json.loads(record.json)
        )

    def to_record(self) -> FilterRecord:
        return FilterRecord(
            name=self.name,
            description=self.description,
            json=self.to_json(),
        )

    def to_file(self) -> str:
        output_json: Dict[str, Any] = {
            attribute: value
            for attribute, value in self.__dict__.items()
            if value and value != ["%"]
        }
        return json.dumps(output_json, indent=4)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-unsafe

import datetime
import logging
import sys
import time
from contextlib import contextmanager
from functools import wraps
from typing import Any, Callable, List, Optional

log = logging.getLogger("sapp")


class retryable(object):
    def __init__(
        self, num_tries: int = 1, retryable_exs: Optional[List[Any]] = None
    ) -> None:
        self.num_tries = num_tries
        self.retryable_exs = retryable_exs

    def __call__(self, func):
        @wraps(func)
        def new_func(*args, **kwargs):
            try_num = 1
            while True:
                try:
                    return func(*args, **kwargs)
                except Exception as e:
                    if self.retryable_exs and type(e) not in self.retryable_exs:
                        raise
                    try_num += 1
                    if try_num > self.num_tries:
                        raise

        new_func.__wrapped__ = func
        return new_func


def log_time(func: Callable[..., Any]) -> Callable[..., Any]:
    """Log the time it takes to run a function. It's sort of like timeit, but
    prettier.
    """

    def wrapper(*args, **kwargs):
        start_time = time.time()
        log.info("%s starting...", func.__name__.title())
        ret = func(*args, **kwargs)
        log.info(
            "%s finished (%s)",
            func.__name__.title(),
            datetime.timedelta(seconds=int(time.time() - start_time)),
        )
        return ret

    return wrapper


class UserError(Exception):
    pass


@contextmanager
def catch_user_error():
    try:
        yield
    except UserError as error:
        print(str(error), file=sys.stderr)


@contextmanager
def catch_keyboard_interrupt():
    try:
        yield
    except KeyboardInterrupt:
        print("\nOperation aborted.", file=sys.stderr)


# For use on enums to alias upper case value.
#
# FLAKE8 does not understand that this is a static property
# flake8: noqa B902
class classproperty(property):
    def __get__(self, cls, owner):
        return classmethod(self.fget).__get__(None, owner)()
